"""
References
-----------

    .. [CullenDehnen2010] Cullen, Lee, and Walter Dehnen. “Inviscid Smoothed
        Particle Hydrodynamics: Inviscid Smoothed Particle Hydrodynamics.”
        Monthly Notices of the Royal Astronomical Society 408, no. 2
        (October 21, 2010): 669–83.
        https://doi.org/10.1111/j.1365-2966.2010.17158.x.

    .. [ReadHayfield2012] Read, J. I., and T. Hayfield. “SPHS: Smoothed
        Particle Hydrodynamics with a Higher Order Dissipation Switch:
        SPH with a Higher Order Dissipation Switch.” Monthly Notices of the
        Royal Astronomical Society 422, no. 4 (June 1, 2012): 3037–55.
        https://doi.org/10.1111/j.1365-2966.2012.20819.x.
    """
from compyle.types import declare
from pysph.base.particle_array import get_ghost_tag

from pysph.sph.equation import Equation
from pysph.sph.integrator_step import IntegratorStep
from pysph.sph.scheme import Scheme
from pysph.sph.wc.linalg import augmented_matrix, gj_solve, identity, mat_mult

GHOST_TAG = get_ghost_tag()


class PSPHScheme(Scheme):
    def __init__(self, fluids, solids, dim, gamma, hfact, betab=2.0, fkern=1.0,
                 max_density_iterations=250, alphac=0.25,
                 density_iteration_tolerance=1e-3, has_ghosts=False,
                 alphamin=0.02, alphamax=2.0, betac=0.7, betad=0.05,
                 betaxi=1.0):
        """
        Pressure-energy formulation [Hopkins2013]_ including Cullen-Dehnen
        artificial viscocity switch [CullenDehnen2010]_ with modifications,
        as presented in Appendix F2 of [Hopkins2015]_ .

        Notes
        -----
        Is this exactly in accordance with what is proposed in [Hopkins2015]_ ?
            Not quite.

        What is different then?
            #. Adapting smoothing length using MPM [KP14]_ procedure from
               :class:`SummationDensity
               <pysph.sph.gas_dynamics.basic.SummationDensity>`. In this,
               calculation of grad-h terms are changed to that specified for
               this scheme.
            #. Using the PEC integrator step. No individual
               adaptive time-stepping.
            #. Using :class:`Gaussian Kernel <pysph.base.kernels.Gaussian>`
               by default instead of Cubic Spline with radius scale 1.

        Tip: Reduce the number of points if particle penetration is
        encountered. This has to be done while running
        ``gas_dynamics.wc_blastwave`` and ``gas_dynamics.robert``.

        Parameters
        ----------
        fluids: list
            List of names of fluid particle arrays.
        solids: list
            List of names of solid particle arrays (or boundaries), currently
            not supported
        dim: int
            Dimensionality of the problem.
        gamma: float
            :math:`\\gamma` for Equation of state.
        hfact: float
            :math:`h_{fact}` for smoothing length adaptivity, also referred to
            as kernel_factor in other gas dynamics schemes.
        betab : float, optional
            :math:`\\beta_b` for artificial viscosity, by default 2.0
        fkern : float, optional
            :math:`f_{kern}`, Factor to scale smoothing length for equivalence
            with classic kernel when using kernel with altered
            `radius_scale` is being used, by default 1.
        max_density_iterations : int, optional
            Maximum number of iterations to run for one density step,
            by default 250.
        alphac : float, optional
            :math:`\\alpha_c` for artificial conductivity, by default 0.25
        density_iteration_tolerance : float, optional
            Maximum difference allowed in two successive density iterations,
            by default 1e-3
        has_ghosts : bool, optional
            if ghost particles (either mirror or periodic) is used, by default
            False
        alphamin : float, optional
            :math:`\\alpha_{min}` for artificial viscosity switch, by default
            0.02
        alphamax : float, optional
            :math:`\\alpha_{max}` for artificial viscosity switch, by default
            2.0
        betac : float, optional
            :math:`\\beta_c` for artificial viscosity switch, by default 0.7
        betad : float, optional
            :math:`\\beta_d` for artificial viscosity switch, by default 0.05
        betaxi : float, optional
            :math:`\\beta_{\\xi}` for artificial viscosity switch,
            by default 1.0
        """

        self.fluids = fluids
        self.solids = solids
        self.dim = dim
        self.solver = None
        self.gamma = gamma
        self.betab = betab
        self.hfact = hfact
        self.density_iteration_tolerance = density_iteration_tolerance
        self.max_density_iterations = max_density_iterations
        self.has_ghosts = has_ghosts
        self.fkern = fkern
        self.alphac = alphac
        self.alphamin = alphamin
        self.alphamax = alphamax
        self.betac = betac
        self.betad = betad
        self.betaxi = betaxi

    def add_user_options(self, group):
        group.add_argument("--alpha-max", action="store", type=float,
                           dest="alphamax", default=None,
                           help="alpha_max for artificial viscosity switch. ")

        group.add_argument("--alpha-min", action="store", type=float,
                           dest="alphamin", default=None,
                           help="alpha_min for artificial viscosity switch. ")

        group.add_argument("--beta-b", action="store", type=float,
                           dest="betab", default=None,
                           help="beta for the artificial viscosity.")

        group.add_argument("--beta-xi", action="store", type=float,
                           dest="betaxi", default=None,
                           help="beta_xi for artificial viscosity switch.")

        group.add_argument("--beta-d", action="store", type=float,
                           dest="betad", default=None,
                           help="beta_d for artificial viscosity switch.")

        group.add_argument("--beta-c", action="store", type=float,
                           dest="betac", default=None,
                           help="beta_c for artificial viscosity switch.")

        group.add_argument("--alpha-c", action="store", type=float,
                           dest="alphac", default=None,
                           help="alpha_c for artificial conductivity. ")

        group.add_argument("--gamma", action="store", type=float, dest="gamma",
                           default=None, help="gamma for the state equation.")

    def consume_user_options(self, options):
        vars = ['gamma', 'alphamax', 'alphamin', 'alphac', 'betab', 'betaxi',
                'betad', 'betac']
        data = dict((var, self._smart_getattr(options, var)) for var in vars)
        self.configure(**data)

    def configure_solver(self, kernel=None, integrator_cls=None,
                         extra_steppers=None, **kw):

        from pysph.base.kernels import Gaussian
        if kernel is None:
            kernel = Gaussian(dim=self.dim)

        if hasattr(kernel, 'fkern'):
            self.fkern = kernel.fkern
        else:
            self.fkern = 1.0

        steppers = {}
        if extra_steppers is not None:
            steppers.update(extra_steppers)

        from pysph.sph.integrator import PECIntegrator

        cls = integrator_cls if integrator_cls is not None else PECIntegrator
        step_cls = PECStep
        for name in self.fluids:
            if name not in steppers:
                steppers[name] = step_cls()

        integrator = cls(**steppers)

        from pysph.solver.solver import Solver
        self.solver = Solver(dim=self.dim, integrator=integrator,
                             kernel=kernel, **kw)

    def get_equations(self):
        from pysph.sph.equation import Group

        equations = []
        # Find the optimal 'h'
        all_pa = self.fluids + self.solids

        g1 = []
        for fluid in self.fluids:
            g1.append(PSPHSummationDensityAndPressure(
                dest=fluid, sources=all_pa, hfact=self.hfact,
                density_iterations=True, dim=self.dim,
                htol=self.density_iteration_tolerance, gamma=self.gamma))
            equations.append(
                Group(equations=g1, update_nnps=True, iterate=True,
                      max_iterations=self.max_density_iterations))

        g2 = []
        for fluid in self.fluids:
            g2.append(GradientKinsfolkC1(
                dest=fluid, sources=all_pa, dim=self.dim))
            g2.append(SignalVelocity(dest=fluid, sources=all_pa))
        equations.append(Group(equations=g2))

        g3 = []
        for fluid in self.fluids:
            g3.append(LimiterAndAlphas(dest=fluid, sources=all_pa,
                                       alphamin=self.alphamin,
                                       alphamax=self.alphamax,
                                       betac=self.betac, betad=self.betad,
                                       betaxi=self.betaxi, fkern=self.fkern))
        equations.append(Group(equations=g3))

        g4 = []
        for solid in self.solids:
            g4.append(WallBoundary(solid, sources=self.fluids))
        equations.append(Group(equations=g4))

        if self.has_ghosts:
            gh = []
            for fluid in self.fluids:
                gh.append(UpdateGhostProps(dest=fluid, sources=None))
            equations.append(Group(equations=gh, real=False))

        g5 = []
        for fluid in self.fluids:
            g5.append(MomentumAndEnergy(
                dest=fluid, sources=all_pa, dim=self.dim, betab=self.betab,
                fkern=self.fkern, alphac=self.alphac, gamma=self.gamma))
        equations.append(Group(equations=g5))

        return equations

    def setup_properties(self, particles, clean=True):
        import numpy
        particle_arrays = dict([(p.name, p) for p in particles])

        props = ['rho', 'm', 'x', 'y', 'z', 'u', 'v', 'w', 'h', 'cs', 'p', 'e',
                 'au', 'av', 'aw', 'ae', 'pid', 'gid', 'tag', 'dwdh', 'h0',
                 'converged', 'ah', 'arho', 'e0', 'rho0', 'u0', 'v0', 'w0',
                 'x0', 'y0', 'z0']
        more_props = ['drhosumdh', 'n', 'dndh', 'prevn', 'prevdndh',
                      'prevdrhosumdh', 'divv', 'dpsumdh', 'dprevpsumdh', 'an',
                      'adivv', 'trssdsst', 'vsig', 'alpha', 'alpha0', 'xi']
        props.extend(more_props)
        output_props = 'rho p u v w x y z e n divv h alpha'.split(' ')
        for fluid in self.fluids:
            pa = particle_arrays[fluid]
            self._ensure_properties(pa, props, clean)
            pa.add_property('orig_idx', type='int')
            # Guess for number density.
            pa.add_property('n', data=pa.rho / pa.m)
            pa.add_property('gradv', stride=9)
            pa.add_property('invtt', stride=9)
            pa.add_property('grada', stride=9)
            pa.add_property('ss', stride=6)
            nfp = pa.get_number_of_particles()
            pa.orig_idx[:] = numpy.arange(nfp)
            pa.set_output_arrays(output_props)

        solid_props = set(props) | set('m0 wij htmp'.split(' '))
        for solid in self.solids:
            pa = particle_arrays[solid]
            self._ensure_properties(pa, solid_props, clean)
            pa.set_output_arrays(output_props)


class PSPHSummationDensityAndPressure(Equation):
    def __init__(self, dest, sources, dim, gamma, density_iterations=False,
                 iterate_only_once=False, hfact=1.2, htol=1e-6):
        """
        :class:`SummationDensity
        <pysph.sph.gas_dynamics.basic.SummationDensity>` modified to use
        number density for calculation of grad-h terms and to calculate
        pressure and speed of sound as well.

        Ref. Appendix F2 [Hopkins2015]_

        Parameters
        ----------
        density_iterations : bint, optional
            Flag to indicate density iterations are required, by default False
        iterate_only_once : bint, optional
            Flag to indicate if only one iteration is required,
             by default False
        hfact : float, optional
            :math:`h_{fact}`, by default 1.2
        htol : double, optional
            Iteration tolerance, by default 1e-6
        """

        self.density_iterations = density_iterations
        self.iterate_only_once = iterate_only_once
        self.dim = dim
        self.hfact = hfact
        self.htol = htol
        self.equation_has_converged = 1
        self.gamma = gamma
        self.gammam1 = gamma - 1.0

        super().__init__(dest, sources)

    def initialize(self, d_idx, d_rho, d_arho, d_n, d_dndh, d_prevn,
                   d_prevdndh, d_p, d_dpsumdh, d_dprevpsumdh, d_an):

        d_rho[d_idx] = 0.0
        d_arho[d_idx] = 0.0

        d_prevn[d_idx] = d_n[d_idx]
        d_prevdndh[d_idx] = d_dndh[d_idx]
        d_n[d_idx] = 0.0
        d_dndh[d_idx] = 0.0
        d_an[d_idx] = 0.0

        d_p[d_idx] = 0.0
        d_dprevpsumdh[d_idx] = d_dpsumdh[d_idx]
        d_dpsumdh[d_idx] = 0.0

        self.equation_has_converged = 1

    def loop(self, d_idx, s_idx, d_rho, d_arho, s_m, VIJ, WI, DWI, GHI, d_n,
             d_dndh, d_h, d_prevn, d_prevdndh, s_e, d_p, d_dpsumdh, d_e, d_an):

        mj = s_m[s_idx]
        vijdotdwij = VIJ[0] * DWI[0] + VIJ[1] * DWI[1] + VIJ[2] * DWI[2]

        # density
        mj_wi = mj * WI
        d_rho[d_idx] += mj_wi
        d_p[d_idx] += self.gammam1 * s_e[s_idx] * mj_wi

        # number density accelerations
        hibynidim = d_h[d_idx] / (d_prevn[d_idx] * self.dim)
        inbrkti = 1 + d_prevdndh[d_idx] * hibynidim
        inprthsi = d_dpsumdh[d_idx] * hibynidim / (
                self.gammam1 * s_m[s_idx] * d_e[d_idx])
        fij = 1 - inprthsi / inbrkti
        vijdotdwij_fij = vijdotdwij * fij
        d_an[d_idx] += vijdotdwij_fij

        # density acceleration is not essential as such
        d_arho[d_idx] += mj * vijdotdwij_fij

        # gradient of kernel w.r.t h
        d_dpsumdh[d_idx] += mj * self.gammam1 * d_e[d_idx] * GHI
        d_n[d_idx] += WI
        d_dndh[d_idx] += GHI

    def post_loop(self, d_idx, d_rho, d_h0, d_h, d_ah, d_converged, d_cs, d_p,
                  d_n, d_dndh, d_an):

        d_cs[d_idx] = sqrt(self.gamma * d_p[d_idx] / d_rho[d_idx])

        # iteratively find smoothing length consistent with the
        if self.density_iterations:
            if not (d_converged[d_idx] == 1):
                hi = d_h[d_idx]
                hi0 = d_h0[d_idx]

                # estimated, without summations
                ni = (self.hfact / hi) ** self.dim
                dndhi = - self.dim * d_n[d_idx] / hi

                # the non-linear function and it's derivative
                func = d_n[d_idx] - ni
                dfdh = d_dndh[d_idx] - dndhi

                # Newton Raphson estimate for the new h
                hnew = hi - func / dfdh

                # Nanny control for h
                if hnew > 1.2 * hi:
                    hnew = 1.2 * hi
                elif hnew < 0.8 * hi:
                    hnew = 0.8 * hi

                # check for convergence
                diff = abs(hnew - hi) / hi0

                if not ((diff < self.htol) or self.iterate_only_once):
                    # this particle hasn't converged. This means the
                    # entire group must be repeated until this fellow
                    # has converged, or till the maximum iteration has
                    # been reached.
                    self.equation_has_converged = -1

                    # set particle properties for the next
                    # iteration. For the 'converged' array, a value of
                    # 0 indicates the particle hasn't converged
                    d_h[d_idx] = hnew
                    d_converged[d_idx] = 0
                else:
                    d_ah[d_idx] = d_an[d_idx] / dndhi
                    d_converged[d_idx] = 1

    def converged(self):
        return self.equation_has_converged


class GradientKinsfolkC1(Equation):
    def __init__(self, dest, sources, dim):
        """
        First order consistent,

            - Velocity gradient, grad(v)
            - Acceleration gradient, grad(a)
            - Velocity divergence, div(v)
            - Velocity divergence rate, d (div(v)) / dt
            - Traceless symmetric strain rate, S
            - trace(dot(S,transpose(S)))

        Ref. Appendix B [CullenDehnen2010]_
        """
        self.dim = dim
        super().__init__(dest, sources)

    def _get_helpers_(self):
        return [augmented_matrix, gj_solve, identity, mat_mult]

    def initialize(self, d_gradv, d_idx, d_invtt, d_divv, d_grada, d_adivv,
                   d_trssdsst):
        start_indx, i, dim = declare('int', 3)
        start_indx = 9 * d_idx
        for i in range(9):
            d_gradv[start_indx + i] = 0.0
            d_invtt[start_indx + i] = 0.0
            d_grada[start_indx + i] = 0.0

        d_divv[d_idx] = 0.0
        d_adivv[d_idx] = 0.0
        d_trssdsst[d_idx] = 0.0

    def loop(self, d_idx, d_invtt, s_m, s_idx, VIJ, DWI, XIJ, d_gradv, d_grada,
             d_au, s_au, d_av, s_av, d_aw, s_aw):
        start_indx, row, col, drowcol, dim = declare('int', 5)
        dim = self.dim
        start_indx = d_idx * 9

        aij = declare('matrix(3)')

        aij[0] = d_au[d_idx] - s_au[s_idx]
        aij[1] = d_av[d_idx] - s_av[s_idx]
        aij[2] = d_aw[d_idx] - s_aw[s_idx]

        for row in range(dim):
            for col in range(dim):
                drowcol = start_indx + row * 3 + col
                d_invtt[drowcol] -= s_m[s_idx] * XIJ[row] * DWI[col]
                d_gradv[drowcol] -= s_m[s_idx] * VIJ[row] * DWI[col]
                d_grada[drowcol] -= s_m[s_idx] * aij[row] * DWI[col]

    def post_loop(self, d_idx, d_gradv, d_invtt, d_divv, d_grada, d_adivv,
                  d_ss, d_trssdsst):
        tt = declare('matrix(9)')
        invtt = declare('matrix(9)')
        augtt = declare('matrix(18)')
        idmat = declare('matrix(9)')
        gradv = declare('matrix(9)')
        grada = declare('matrix(9)')

        start_indx, row, col, rowcol, drowcol, dim, colrow = declare('int', 7)
        ltstart_indx, dltrowcol = declare('int', 2)
        dim = self.dim
        start_indx = 9 * d_idx
        identity(idmat, 3)
        identity(tt, 3)

        for row in range(3):
            for col in range(3):
                rowcol = row * 3 + col
                drowcol = start_indx + rowcol
                gradv[rowcol] = d_gradv[drowcol]
                grada[rowcol] = d_grada[drowcol]

        for row in range(dim):
            for col in range(dim):
                rowcol = row * 3 + col
                drowcol = start_indx + rowcol
                tt[rowcol] = d_invtt[drowcol]

        augmented_matrix(tt, idmat, 3, 3, 3, augtt)
        gj_solve(augtt, 3, 3, invtt)
        gradvls = declare('matrix(9)')
        gradals = declare('matrix(9)')
        mat_mult(gradv, invtt, 3, gradvls)
        mat_mult(grada, invtt, 3, gradals)

        for row in range(dim):
            d_divv[d_idx] += gradvls[row * 3 + row]
            d_adivv[d_idx] += gradals[row * 3 + row]
            for col in range(dim):
                rowcol = row * 3 + col
                colrow = row + col * 3
                drowcol = start_indx + rowcol
                d_gradv[drowcol] = gradvls[rowcol]
                d_grada[drowcol] = gradals[rowcol]
                d_adivv[d_idx] -= gradals[rowcol] * gradals[colrow]

        # Traceless Symmetric Strain Rate
        divvbydim = d_divv[d_idx] / dim
        start_indx = d_idx * 9
        ltstart_indx = d_idx * 6
        for row in range(dim):
            col = row
            rowcol = start_indx + row * 3 + col
            dltrowcol = ltstart_indx + (row * (row + 1)) / 2 + col
            d_ss[dltrowcol] = d_gradv[rowcol] - divvbydim

        for row in range(1, dim):
            for col in range(row):
                rowcol = row * 3 + col
                colrow = row + col * 3
                dltrowcol = ltstart_indx + (row * (row + 1)) / 2 + col
                d_ss[dltrowcol] = 0.5 * (gradvls[rowcol] + gradvls[colrow])

        # Trace ( S dot transpose(S) )
        for row in range(dim):
            for col in range(dim):
                dltrowcol = ltstart_indx + (row * (row + 1)) / 2 + col
                d_trssdsst[d_idx] += d_ss[dltrowcol] * d_ss[dltrowcol]


class SignalVelocity(Equation):
    """
    Ref. Equation 25 [Hopkins2015]_
    """

    def initialize(self, d_idx, d_vsig):
        d_vsig[d_idx] = 0.0

    def loop_all(self, d_idx, d_x, d_y, d_z, s_x, s_y, s_z, d_u, d_v, d_w, s_u,
                 s_v, s_w, d_cs, s_cs, d_vsig, NBRS, N_NBRS):
        i = declare('int')
        s_idx = declare('long')
        xij = declare('matrix(3)')
        vij = declare('matrix(3)')
        vijdotxij = 0.0
        cij = 0.0

        for i in range(N_NBRS):
            s_idx = NBRS[i]
            xij[0] = d_x[d_idx] - s_x[s_idx]
            xij[1] = d_y[d_idx] - s_y[s_idx]
            xij[2] = d_z[d_idx] - s_z[s_idx]

            vij[0] = d_u[d_idx] - s_u[s_idx]
            vij[1] = d_v[d_idx] - s_v[s_idx]
            vij[2] = d_w[d_idx] - s_w[s_idx]

            vijdotxij = vij[0] * xij[0] + vij[1] * xij[1] + vij[2] * xij[2]
            cij = 0.5 * (d_cs[d_idx] + s_cs[s_idx])

            d_vsig[d_idx] = max(d_vsig[d_idx], cij - min(0, vijdotxij))


class LimiterAndAlphas(Equation):
    def __init__(self, dest, sources, alphamin=0.02, alphamax=2.0, betac=0.7,
                 betad=0.05, betaxi=1.0, fkern=1.0):
        """
        Cullen Dehnen's limiter for artificial viscosity modified by Hopkins.

        Ref. Appendix F2 [Hopkins2015]_
        """
        self.alphamin = alphamin
        self.alphamax = alphamax
        self.betac = betac
        self.betad = betad
        self.betaxi = betaxi
        self.fkern = fkern
        super().__init__(dest, sources)

    def initialize(self, d_idx, d_xi):
        d_xi[d_idx] = 0.0

    def loop(self, d_idx, s_idx, s_m, d_xi, s_divv, WI):

        if s_divv[s_idx] < 0:
            sign = -1.0
        else:
            sign = 1.0

        d_xi[d_idx] += sign * s_m[s_idx] * WI

    def post_loop(self, d_idx, d_xi, d_rho, d_h, d_adivv, d_cs, d_alpha0,
                  d_vsig, dt, d_divv, d_trssdsst, d_alpha):
        d_xi[d_idx] = 1.0 - d_xi[d_idx] / d_rho[d_idx]
        fhi = self.fkern * d_h[d_idx]

        if d_adivv[d_idx] >= 0 or d_divv[d_idx] >= 0:
            alphatmp = 0.0
        else:
            absadivv = abs(d_adivv[d_idx])
            csbyfhi = d_cs[d_idx] / fhi
            alphatmp = self.alphamax * absadivv / (
                    absadivv + self.betac * csbyfhi * csbyfhi)

        if alphatmp >= d_alpha0[d_idx]:
            d_alpha0[d_idx] = alphatmp
        elif alphatmp < d_alpha0[d_idx]:  # Isn't this obvious?
            epow = exp(-self.betad * dt * abs(d_vsig[d_idx]) * 0.5 / fhi)
            d_alpha0[d_idx] = alphatmp + (d_alpha0[d_idx] - alphatmp) * epow

        xip4 = d_xi[d_idx] * d_xi[d_idx] * d_xi[d_idx] * d_xi[d_idx]
        alnumtt = self.betaxi * xip4 * d_divv[d_idx]
        alnumt = alnumtt * alnumtt
        alnum = alnumt * d_alpha0[d_idx]
        alden = alnumt + d_trssdsst[d_idx]

        if alden < 1e-8:
            d_alpha[d_idx] = self.alphamin
        else:
            d_alpha[d_idx] = max(alnum / alden, self.alphamin)


class MomentumAndEnergy(Equation):
    def __init__(self, dest, sources, dim, fkern, gamma, betab=2.0,
                 alphac=0.25):
        r"""
        PSPH Momentum and Energy Equations with artificial viscosity and
        artificial conductivity.

        Possible typos in that have been taken care of,

            1. Instead of Equation F15 [Hopkins2015]_ for evolution of total
               energy sans artificial viscosity and artificial conductivity,

                .. math::
                    \frac{\mathrm{d} E_{i}}{\mathrm{~d} t}= \boldsymbol{v}_{i}
                    \cdot \frac{\mathrm{d} \boldsymbol{P}_{i}}{\mathrm{~d} t}-
                    \sum_{j=1}^{N}(\gamma-1)^{2} m_{i} m_{j} u_{i} u_{j}
                    \frac{f_{i j}}{\bar{P}_{i}}\left(\boldsymbol{v}_{i}-
                    \boldsymbol{v}_{j}\right) \cdot \nabla_{i} W_{i j}
                    \left(h_{i}\right),

               it should have been,

                .. math::
                    \frac{\mathrm{d} E_{i}}{\mathrm{~d} t}= \boldsymbol{v}_{i}
                    \cdot \frac{\mathrm{d} \boldsymbol{P}_{i}}{\mathrm{~d} t}+
                    \sum_{j=1}^{N}(\gamma-1)^{2} m_{i} m_{j} u_{i} u_{j}
                    \frac{f_{i j}}{\bar{P}_{i}}\left(\boldsymbol{v}_{i}-
                    \boldsymbol{v}_{j}\right) \cdot \nabla_{i} W_{i j}
                    \left(h_{i}\right).

               Specific thermal energy, :math:`u`, would therefore be evolved
               using,

                .. math::
                    \frac{\mathrm{d} u_{i}}{\mathrm{~d} t}=
                    \sum_{j=1}^{N}(\gamma-1)^{2} m_{j} u_{i} u_{j}
                    \frac{f_{i j}}{\bar{P}_{i}}\left(\boldsymbol{v}_{i}-
                    \boldsymbol{v}_{j}\right) \cdot \nabla_{i} W_{i j}
                    \left(h_{i}\right).

            #. Equation F18 [Hopkins2015]_ for contribution of
               artificial viscosity to the evolution of total
               energy is,

                .. math::
                    \frac{\mathrm{d} E_{i}}{\mathrm{~d} t}= \alpha_{\mathrm{C}}
                    \sum_{j} m_{i} m_{j} \alpha_{i j} \tilde{v}_{s}\left(u_{i}-
                    u_{j}\right) \times \frac{\left|P_{i}-P_{j}\right|}{P_{i}+
                    P_{j}} \frac{\nabla_{i} W_{i j}\left(h_{i}\right)+
                    \nabla_{i} W_{i j}\left(h_{j}\right)}{\bar{\rho}_{i}+
                    \bar{\rho}_{j}} .

               Carefully comparing with [ReadHayfield2012]_ and [KP14]_,
               specific thermal energy, :math:`u`, should be evolved
               using,

                .. math::
                    \frac{\mathrm{d} u_{i}}{\mathrm{~d} t}= \alpha_{\mathrm{C}}
                    \sum_{j} & m_{j} \alpha_{i j} \tilde{v}_{s}\left(u_{i}-
                    u_{j}\right) \frac{\left|P_{i}-P_{j}\right|}{P_{i}+
                    P_{j}} \\ & \frac{\nabla_{i} W_{i j}\left(h_{i}\right)+
                    \nabla_{i} W_{i j}\left(h_{j}\right)}{\bar{\rho}_{i}+
                    \bar{\rho}_{j}} \cdot \frac{\left(\boldsymbol{x}_{i}-
                    \boldsymbol{x}_{j}\right)}{\left|\boldsymbol{x}_{i}-
                    \boldsymbol{x}_{j}\right|}
        """
        self.betab = betab
        self.dim = dim
        self.fkern = fkern
        self.alphac = alphac
        self.gammam1 = gamma - 1.0
        super().__init__(dest, sources)

    def initialize(self, d_idx, d_au, d_av, d_aw, d_ae):
        d_au[d_idx] = 0.0
        d_av[d_idx] = 0.0
        d_aw[d_idx] = 0.0
        d_ae[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_m, s_m, d_p, s_p, d_cs, s_cs, d_au, d_av,
             d_aw, d_ae, XIJ, VIJ, DWI, DWJ, d_alpha, s_alpha, RIJ, d_h,
             d_dndh, d_n, s_h, s_dndh, s_n, d_e, s_e, d_dpsumdh, s_dpsumdh,
             RHOIJ1):

        dim = self.dim
        gammam1 = self.gammam1
        avi = declare("matrix(3)")

        # averaged sound speed
        cij = 0.5 * (d_cs[d_idx] + s_cs[s_idx])

        mj = s_m[s_idx]
        vijdotxij = (VIJ[0] * XIJ[0] + VIJ[1] * XIJ[1] + VIJ[2] * XIJ[2])

        if RIJ < 1e-8:
            vs = 2 * cij
            muij = 0.0
            Fij = 0.0
        else:
            vs = 2 * cij - 3 * vijdotxij / RIJ
            muij = vijdotxij / RIJ
            Fij = 0.5 * (XIJ[0] * (DWI[0] + DWJ[0]) +
                         XIJ[1] * (DWI[1] + DWJ[1]) +
                         XIJ[2] * (DWI[2] + DWJ[2])) / RIJ

        # Artificial viscosity
        if vijdotxij <= 0.0:
            alphaij = 0.5 * (d_alpha[d_idx] + s_alpha[s_idx])
            oby2rhoij = RHOIJ1 / 2.0
            common = (alphaij * muij * (
                    cij - self.betab * muij) * mj * oby2rhoij)

            avi[0] = common * (DWI[0] + DWJ[0])
            avi[1] = common * (DWI[1] + DWJ[1])
            avi[2] = common * (DWI[2] + DWJ[2])

            # viscous contribution to velocity
            d_au[d_idx] += avi[0]
            d_av[d_idx] += avi[1]
            d_aw[d_idx] += avi[2]

            # viscous contribution to the thermal energy
            d_ae[d_idx] -= 0.5 * (VIJ[0] * avi[0] +
                                  VIJ[1] * avi[1] +
                                  VIJ[2] * avi[2])

            # artificial conductivity
            eij = d_e[d_idx] - s_e[s_idx]
            Lij = abs(d_p[d_idx] - s_p[s_idx]) / (d_p[d_idx] + s_p[s_idx])
            d_ae[d_idx] += (self.alphac * mj * alphaij * vs * eij * Lij * Fij *
                            oby2rhoij)

        # grad-h correction terms.
        hibynidim = d_h[d_idx] / (d_n[d_idx] * dim)
        inbrkti = 1 + d_dndh[d_idx] * hibynidim
        inprthsi = d_dpsumdh[d_idx] * hibynidim / (
                gammam1 * s_m[s_idx] * d_e[d_idx])
        fij = 1 - inprthsi / inbrkti

        hjbynjdim = s_h[s_idx] / (s_n[s_idx] * dim)
        inbrktj = 1 + s_dndh[s_idx] * hjbynjdim
        inprthsj = s_dpsumdh[s_idx] * hjbynjdim / (
                gammam1 * d_m[d_idx] * s_e[s_idx])
        fji = 1 - inprthsj / inbrktj

        # accelerations for velocity
        gammam1sq = gammam1 * gammam1
        comm = gammam1sq * mj * d_e[d_idx] * s_e[s_idx]
        commi = comm * fij / d_p[d_idx]
        commj = comm * fji / s_p[s_idx]

        d_au[d_idx] -= commi * DWI[0] + commj * DWJ[0]
        d_av[d_idx] -= commi * DWI[1] + commj * DWJ[1]
        d_aw[d_idx] -= commi * DWI[2] + commj * DWJ[2]

        # accelerations for the thermal energy
        vijdotdwi = VIJ[0] * DWI[0] + VIJ[1] * DWI[1] + VIJ[2] * DWI[2]
        d_ae[d_idx] += commi * vijdotdwi


class WallBoundary(Equation):
    """
        :class:`WallBoundary
        <pysph.sph.gas_dynamics.boundary_equations.WallBoundary>` modified
        for PSPH
    """

    def initialize(self, d_idx, d_p, d_rho, d_e, d_m, d_cs, d_h, d_htmp, d_h0,
                   d_u, d_v, d_w, d_wij, d_n, d_dndh, d_dpsumdh, d_m0):

        d_p[d_idx] = 0.0
        d_u[d_idx] = 0.0
        d_v[d_idx] = 0.0
        d_w[d_idx] = 0.0
        d_m0[d_idx] = d_m[d_idx]
        d_m[d_idx] = 0.0
        d_rho[d_idx] = 0.0
        d_e[d_idx] = 0.0
        d_cs[d_idx] = 0.0
        d_wij[d_idx] = 0.0
        d_h[d_idx] = d_h0[d_idx]
        d_htmp[d_idx] = 0.0
        d_n[d_idx] = 0.0
        d_dndh[d_idx] = 0.0
        d_dpsumdh[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_p, d_rho, d_e, d_m, d_cs, d_u, d_v, d_w,
             d_wij, d_htmp, s_p, s_rho, s_e, s_m, s_cs, s_h, s_u, s_v, s_w, WI,
             s_n, d_n, d_dndh, s_dndh, d_dpsumdh, s_dpsumdh):
        d_wij[d_idx] += WI
        d_p[d_idx] += s_p[s_idx] * WI
        d_u[d_idx] -= s_u[s_idx] * WI
        d_v[d_idx] -= s_v[s_idx] * WI
        d_w[d_idx] -= s_w[s_idx] * WI
        d_m[d_idx] += s_m[s_idx] * WI
        d_rho[d_idx] += s_rho[s_idx] * WI
        d_e[d_idx] += s_e[s_idx] * WI
        d_cs[d_idx] += s_cs[s_idx] * WI
        d_htmp[d_idx] += s_h[s_idx] * WI
        d_n[d_idx] += s_n[s_idx] * WI
        d_dndh[d_idx] += s_dndh[s_idx] * WI
        d_dpsumdh[d_idx] += s_dpsumdh[s_idx] * WI

    def post_loop(self, d_idx, d_p, d_rho, d_e, d_m, d_cs, d_h, d_u, d_v, d_w,
                  d_wij, d_htmp, d_dndh, d_dpsumdh, d_n, d_m0):
        if d_wij[d_idx] > 1e-30:
            d_p[d_idx] = d_p[d_idx] / d_wij[d_idx]
            d_u[d_idx] = d_u[d_idx] / d_wij[d_idx]
            d_v[d_idx] = d_v[d_idx] / d_wij[d_idx]
            d_w[d_idx] = d_w[d_idx] / d_wij[d_idx]
            d_m[d_idx] = d_m[d_idx] / d_wij[d_idx]
            d_rho[d_idx] = d_rho[d_idx] / d_wij[d_idx]
            d_e[d_idx] = d_e[d_idx] / d_wij[d_idx]
            d_cs[d_idx] = d_cs[d_idx] / d_wij[d_idx]
            d_h[d_idx] = d_htmp[d_idx] / d_wij[d_idx]
            d_dndh[d_idx] /= d_wij[d_idx]
            d_dpsumdh[d_idx] /= d_wij[d_idx]
            d_n[d_idx] /= d_wij[d_idx]

        # Secret Sauce
        if d_m[d_idx] < 1e-10:
            d_m[d_idx] = d_m0[d_idx]


class UpdateGhostProps(Equation):
    def __init__(self, dest, sources=None, dim=2):
        """
        :class:`MPMUpdateGhostProps
        <pysph.sph.gas_dynamics.basic.MPMUpdateGhostProps>` modified
        for PSPH
        """
        super().__init__(dest, sources)
        self.dim = dim
        assert GHOST_TAG == 2

    def initialize(self, d_idx, d_orig_idx, d_p, d_tag, d_h, d_rho, d_dndh,
                   d_psumdh, d_n):
        idx = declare('int')
        if d_tag[d_idx] == 2:
            idx = d_orig_idx[d_idx]
            d_p[d_idx] = d_p[idx]
            d_h[d_idx] = d_h[idx]
            d_rho[d_idx] = d_rho[idx]
            d_dndh[d_idx] = d_dndh[idx]
            d_psumdh[d_idx] = d_psumdh[idx]
            d_n[d_idx] = d_n[idx]


class PECStep(IntegratorStep):
    """Predictor Corrector integrator for Gas-dynamics modified for PSPH"""

    def initialize(self, d_idx, d_x0, d_y0, d_z0, d_x, d_y, d_z, d_h, d_u0,
                   d_v0, d_w0, d_u, d_v, d_w, d_e, d_e0, d_h0, d_converged,
                   d_rho, d_rho0):
        d_x0[d_idx] = d_x[d_idx]
        d_y0[d_idx] = d_y[d_idx]
        d_z0[d_idx] = d_z[d_idx]

        d_u0[d_idx] = d_u[d_idx]
        d_v0[d_idx] = d_v[d_idx]
        d_w0[d_idx] = d_w[d_idx]

        d_e0[d_idx] = d_e[d_idx]

        d_h0[d_idx] = d_h[d_idx]
        d_rho0[d_idx] = d_rho[d_idx]

        # set the converged attribute to 0 at the beginning of a Group
        d_converged[d_idx] = 0

    def stage1(self, d_idx, d_x0, d_y0, d_z0, d_x, d_y, d_z, d_u0, d_v0, d_w0,
               d_u, d_v, d_w, d_e0, d_e, d_au, d_av, d_aw, d_ae, d_rho, d_rho0,
               d_arho, d_h, d_h0, d_ah, dt):
        dtb2 = 0.5 * dt

        d_u[d_idx] = d_u0[d_idx] + dtb2 * d_au[d_idx]
        d_v[d_idx] = d_v0[d_idx] + dtb2 * d_av[d_idx]
        d_w[d_idx] = d_w0[d_idx] + dtb2 * d_aw[d_idx]

        d_x[d_idx] = d_x0[d_idx] + dtb2 * d_u[d_idx]
        d_y[d_idx] = d_y0[d_idx] + dtb2 * d_v[d_idx]
        d_z[d_idx] = d_z0[d_idx] + dtb2 * d_w[d_idx]

        # update thermal energy
        d_e[d_idx] = d_e0[d_idx] + dtb2 * d_ae[d_idx]

        # predict density and smoothing lengths for faster
        # convergence. NNPS need not be explicitly updated since it
        # will be called at the end of the predictor stage.
        d_h[d_idx] = d_h0[d_idx] + dtb2 * d_ah[d_idx]
        d_rho[d_idx] = d_rho0[d_idx] + dtb2 * d_arho[d_idx]

    def stage2(self, d_idx, d_x0, d_y0, d_z0, d_x, d_y, d_z, d_u0, d_v0, d_w0,
               d_u, d_v, d_w, d_e0, d_e, d_au, d_av, d_aw, d_ae, dt):
        d_u[d_idx] = d_u0[d_idx] + dt * d_au[d_idx]
        d_v[d_idx] = d_v0[d_idx] + dt * d_av[d_idx]
        d_w[d_idx] = d_w0[d_idx] + dt * d_aw[d_idx]

        d_x[d_idx] = d_x0[d_idx] + dt * d_u[d_idx]
        d_y[d_idx] = d_y0[d_idx] + dt * d_v[d_idx]
        d_z[d_idx] = d_z0[d_idx] + dt * d_w[d_idx]

        # Update densities and smoothing lengths from the accelerations
        d_e[d_idx] = d_e0[d_idx] + dt * d_ae[d_idx]



"""Example for the Noh's cylindrical implosion test. (10 minutes)
"""

# NumPy and standard library imports
import numpy

# PySPH base and carray imports
from pysph.base.utils import get_particle_array as gpa
from pysph.solver.application import Application
from pysph.sph.scheme import GasDScheme, SchemeChooser, ADKEScheme, GSPHScheme
from pysph.sph.gas_dynamics.psph import PSPHScheme
from pysph.sph.gas_dynamics.tsph import TSPHScheme
from pysph.sph.wc.crksph import CRKSPHScheme
from pysph.base.nnps import DomainManager

# problem constants
dim = 2
gamma = 5.0/3.0
gamma1 = gamma - 1.0

# scheme constants
alpha1 = 1.0
alpha2 = 0.1
beta = 2.0
kernel_factor = 1.5

# numerical constants
dt = 1e-3
tf = 0.6

# domain and particle spacings
xmin = ymin = -1.0
xmax = ymax = 1.0

nx = ny = 100
dx = (xmax-xmin)/nx
dxb2 = 0.5 * dx

# initial values
h0 = kernel_factor*dx
rho0 = 1.0
m0 = dx*dx * rho0
vr = -1.0


class NohImplosion(Application):
    def create_particles(self):
        x, y = numpy.mgrid[
            xmin:xmax:dx, ymin:ymax:dx]

        # positions
        x = x.ravel()
        y = y.ravel()

        rho = numpy.ones_like(x) * rho0
        m = numpy.ones_like(x) * m0
        h = numpy.ones_like(x) * h0

        u = numpy.ones_like(x)
        v = numpy.ones_like(x)

        sin, cos, arctan = numpy.sin, numpy.cos, numpy.arctan2
        for i in range(x.size):
            theta = arctan(y[i], x[i])
            u[i] = vr*cos(theta)
            v[i] = vr*sin(theta)

        fluid = gpa(
            name='fluid', x=x, y=y, m=m, rho=rho, h=h, u=u, v=v, p=1e-12,
            e=2.5e-11, h0=h.copy()
            )
        self.scheme.setup_properties([fluid])

        print("Noh's problem with %d particles "
              % (fluid.get_number_of_particles()))

        return [fluid, ]

    def create_domain(self):
        return DomainManager(
            xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax,
            mirror_in_x=True, mirror_in_y=True
        )

    def create_scheme(self):
        mpm = GasDScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            kernel_factor=kernel_factor, alpha1=alpha1, alpha2=alpha2,
            beta=beta, adaptive_h_scheme="mpm",
            update_alpha1=True, update_alpha2=True, has_ghosts=True
        )

        crksph = CRKSPHScheme(
            fluids=['fluid'], dim=2, rho0=0, c0=0, nu=0, h0=0, p0=0,
            gamma=gamma, cl=2, has_ghosts=True
        )

        gsph = GSPHScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            kernel_factor=1.5,
            g1=0.25, g2=0.5, rsolver=7, interpolation=1, monotonicity=2,
            interface_zero=True, hybrid=False, blend_alpha=2.0,
            niter=40, tol=1e-6, has_ghosts=True
        )

        adke = ADKEScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            alpha=1, beta=1, k=1.0, eps=0.8, g1=0.5, g2=0.5,
            has_ghosts=True
        )

        psph = PSPHScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            hfact=1.2
        )

        tsph = TSPHScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            hfact=1.2
        )

        s = SchemeChooser(
            default='crksph', crksph=crksph, mpm=mpm, adke=adke, gsph=gsph,
            psph=psph, tsph=tsph
        )
        s.configure_solver(dt=dt, tf=tf, adaptive_timestep=False)
        return s

    def configure_scheme(self):
        s = self.scheme
        if self.options.scheme == 'mpm':
            s.configure(kernel_factor=1.2)
            s.configure_solver(
                dt=dt, tf=tf, adaptive_timestep=True, pfreq=50
            )
        elif self.options.scheme == 'crksph':
            s.configure_solver(
                dt=dt, tf=tf, adaptive_timestep=False, pfreq=50
            )
        elif self.options.scheme == 'gsph':
            s.configure_solver(
                dt=dt, tf=tf, adaptive_timestep=False, pfreq=50
            )
        elif self.options.scheme == 'adke':
            s.configure_solver(
                dt=dt, tf=tf, adaptive_timestep=False, pfreq=50
            )
        elif self.options.scheme in ['tsph', 'psph']:
            s.configure(hfact=1.2)
            s.configure_solver(
                dt=dt, tf=tf, adaptive_timestep=False, pfreq=50
            )

    def post_process(self):
        try:
            import matplotlib
            matplotlib.use('Agg')
            from matplotlib import pyplot
        except ImportError:
            print("Post processing requires matplotlib.")
            return

        if self.rank > 0 or len(self.output_files) == 0:
            return

        import os
        from pysph.solver.utils import load
        outfile = self.output_files[-1]
        data = load(outfile)
        pa = data['arrays']['fluid']

        x = pa.x
        y = pa.y
        rho = pa.rho
        p = pa.p

        r = numpy.sqrt(x**2 + y**2)

        # exact solutions
        vs = 1.0/3.0  # shock radial velocity
        rs = vs * tf  # position of shock
        ri = numpy.linspace(0, rs, 10)
        ro = numpy.linspace(rs, xmax, 100)
        re = numpy.concatenate((ri, ro))

        rho_e1 = numpy.ones_like(ri) * ((gamma + 1) / (gamma - 1))**dim
        rho_e2 = rho0 * (1 + tf / ro)**(dim - 1)
        rho_e = numpy.concatenate((rho_e1, rho_e2))

        p_e1 = vs * rho_e1
        p_e2 = numpy.zeros_like(ro)
        p_e = numpy.concatenate((p_e1, p_e2))

        pyplot.scatter(r, p, s=1)
        pyplot.xlabel('r')
        pyplot.ylabel('P')
        pyplot.plot(re, p_e, color='r', lw=1)
        pyplot.legend(
            ['exact', self.options.scheme]
        )
        fname = os.path.join(self.output_dir, 'pressure.png')
        pyplot.savefig(fname, dpi=300)
        pyplot.close('all')

        pyplot.scatter(r, rho, s=1)
        pyplot.xlabel('r')
        pyplot.ylabel(r'$\rho$')
        pyplot.plot(re, rho_e, color='r', lw=1)
        pyplot.legend(
            ['exact', self.options.scheme]
        )
        fname = os.path.join(self.output_dir, 'density.png')
        pyplot.savefig(fname, dpi=300)
        pyplot.close('all')

        fname = os.path.join(self.output_dir, '1dresults.npz')
        numpy.savez(fname, r=r, rho=rho, p=p)

        fname = os.path.join(self.output_dir, '1dexact.npz')
        numpy.savez(fname, r=re, rho=rho_e, p=p_e)


if __name__ == '__main__':
    app = NohImplosion()
    app.run()
    app.post_process()



from pysph.sph.equation import Equation


class Continuity(Equation):

    def initialize(self, d_idx, d_arho):
        d_arho[d_idx] = 0.0

    def loop(self, d_idx, s_idx, s_m, d_u, d_v, d_w, s_u, s_v, s_w, d_cs, s_cs,
             d_rho, d_arho, s_rho, d_p, s_p, DWIJ, RIJ, XIJ):
        rl = d_rho[d_idx]
        rr = s_rho[s_idx]
        pl = d_p[d_idx]
        pr = s_p[s_idx]
        cl = d_cs[d_idx]
        cr = s_cs[s_idx]
        uxl = d_u[d_idx]
        uyl = d_v[d_idx]
        uzl = d_w[d_idx]
        uxr = s_u[s_idx]
        uyr = s_v[s_idx]
        uzr = s_w[s_idx]

        if RIJ >= 1.0e-16:
            ul = -(uxl * XIJ[0] + uyl * XIJ[1] + uzl * XIJ[2]) / RIJ
            ur = -(uxr * XIJ[0] + uyr * XIJ[1] + uzr * XIJ[2]) / RIJ
        else:
            ul = 0.0
            ur = 0.0
        u_star = (ul * rl * cl + ur * rr * cr + pl - pr) / (rl * cl + rr * cr)
        dwdr = sqrt(DWIJ[0] * DWIJ[0] + DWIJ[1] * DWIJ[1] + DWIJ[2] * DWIJ[2])

        d_arho[d_idx] += 2.0 * s_m[s_idx] * dwdr * (ul - u_star) * rl / rr


class Momentum(Equation):

    def __init__(self, dest, sources, gx=0.0, gy=0.0, gz=0.0):
        self.gx = gx
        self.gy = gy
        self.gz = gz

        super(Momentum, self).__init__(dest, sources)

    def initialize(self, d_idx, d_au, d_av, d_aw):
        d_au[d_idx] = self.gx
        d_av[d_idx] = self.gy
        d_aw[d_idx] = self.gz

    def loop(self, d_idx, s_idx, s_m, d_u, d_v, d_w, s_u, s_v, s_w, d_cs, s_cs,
             d_rho, s_rho, d_p, s_p, d_au, d_av, d_aw, RIJ, XIJ, DWIJ):
        rl = d_rho[d_idx]
        rr = s_rho[s_idx]
        pl = d_p[d_idx]
        pr = s_p[s_idx]
        cl = d_cs[d_idx]
        cr = s_cs[s_idx]
        uxl = d_u[d_idx]
        uyl = d_v[d_idx]
        uzl = d_w[d_idx]
        uxr = s_u[s_idx]
        uyr = s_v[s_idx]
        uzr = s_w[s_idx]
        m = s_m[s_idx]

        if RIJ >= 1.0e-16:
            ul = -(uxl * XIJ[0] + uyl * XIJ[1] + uzl * XIJ[2]) / RIJ
            ur = -(uxr * XIJ[0] + uyr * XIJ[1] + uzr * XIJ[2]) / RIJ
        else:
            ul = 0.0
            ur = 0.0
        p_star = pl * rr * cr + pr * cl * rl - rl * rr * cl * cr * (ur - ul)
        p_star /= (rl * cl + rr * cr)
        factor = -2.0 * m * p_star / (rl * rr)

        d_au[d_idx] += factor * DWIJ[0]
        d_av[d_idx] += factor * DWIJ[1]
        d_aw[d_idx] += factor * DWIJ[2]



"""Two square blocks of water colliding with each other. (20 seconds)
"""

import numpy
from pysph.examples._db_geometry import create_2D_filled_region
from pysph.base.utils import get_particle_array
from pysph.sph.iisph import IISPHScheme
from pysph.solver.application import Application

dx = 0.025
hdx = 1.0
rho0 = 1000


class TwoBlocks(Application):
    def create_particles(self):
        x1, y1 = create_2D_filled_region(-1, 0, 0, 1, dx)
        x2, y2 = create_2D_filled_region(0.5, 0, 1.5, 1, dx)

        x = numpy.concatenate((x1, x2))
        y = numpy.concatenate((y1, y2))
        u1 = numpy.ones_like(x1)
        u2 = -numpy.ones_like(x2)
        u = numpy.concatenate((u1, u2))

        rho = numpy.ones_like(u)*rho0
        h = numpy.ones_like(u)*hdx*dx
        m = numpy.ones_like(u)*dx*dx*rho0

        fluid = get_particle_array(
            name='fluid', x=x, y=y, u=u, rho=rho, m=m, h=h
        )
        self.scheme.setup_properties([fluid])
        return [fluid]

    def create_scheme(self):
        s = IISPHScheme(fluids=['fluid'], solids=[], dim=2, rho0=rho0)
        return s

    def configure_scheme(self):
        dt = 2e-3
        tf = 1.0
        self.scheme.configure_solver(
            dt=dt, tf=tf, adaptive_timestep=False, pfreq=10
        )


if __name__ == '__main__':
    app = TwoBlocks()
    app.run()



# cython: language_level=3, embedsignature=True
# distutils: language=c++
from libcpp.map cimport map
from libcpp.pair cimport pair

from pysph.base.gpu_nnps_base cimport *

ctypedef unsigned int u_int
ctypedef map[u_int, pair[u_int, u_int]] key_to_idx_t

cdef extern from "math.h":
    double log2(double) nogil

cdef extern from "z_order.h":
    ctypedef unsigned long long uint64_t
    ctypedef unsigned int uint32_t

cdef class ZOrderGPUNNPS(GPUNNPS):
    cdef NNPSParticleArrayWrapper src, dst # Current source and destination.

    cdef public list pids
    cdef public list pid_keys
    cdef public list cids
    cdef public list cid_to_idx
    cdef public list max_cid
    cdef public list cell_lengths
    cdef public list cell_start_indices
    cdef public object dst_to_src
    cdef object overflow_cid_to_idx
    cdef object curr_cid
    cdef object max_cid_src
    cdef object allocator

    cdef object helper
    cdef object radix_sort
    cdef object make_vec

    cdef public bint sorted
    cdef bint dst_src

    cdef object z_order_nbrs
    cdef object z_order_nbr_lengths

    #cpdef get_spatially_ordered_indices(self, int pa_index)

    cpdef _bin(self, int pa_index)

    cpdef _refresh(self)

    cdef void find_neighbor_lengths(self, nbr_lengths)

    cdef void find_nearest_neighbors_gpu(self, nbrs, start_indices)



import numpy as np
from unittest import TestCase, main

from pysph.base.reduce_array import serial_reduce_array, dummy_reduce_array


class TestSerialReduceArray(TestCase):
    def test_reduce_sum_works(self):
        x = np.linspace(0, 10, 100)
        expect = np.sum(x)
        result = serial_reduce_array(x, 'sum')
        self.assertAlmostEqual(result, expect)

    def test_reduce_prod_works(self):
        x = np.linspace(0, 10, 100)
        expect = np.prod(x)
        result = serial_reduce_array(x, 'prod')
        self.assertAlmostEqual(result, expect)

    def test_reduce_max_works(self):
        x = np.linspace(0, 10, 100)
        expect = np.max(x)
        result = serial_reduce_array(x, 'max')
        self.assertAlmostEqual(result, expect)

    def test_reduce_min_works(self):
        x = np.linspace(0, 10, 100)
        expect = np.min(x)
        result = serial_reduce_array(x, 'min')
        self.assertAlmostEqual(result, expect)

    def test_reduce_raises_error_for_wrong_op(self):
        x = np.linspace(0, 10, 100)
        self.assertRaises(RuntimeError, serial_reduce_array, x, 'foo')

    def test_dummy_reduce_array_does_nothing(self):
        x = np.array([1.0, 2.0])
        expect = x
        result = dummy_reduce_array(x, 'min')
        self.assertTrue(np.alltrue(result == expect))


if __name__ == '__main__':
    main()



"""Test the ParticleArrayExchange object with the following data

25 particles are created on a rectangular grid with the following
processor assignment:

2---- 3---- 3---- 3---- 3
|     |     |     |     |
1---- 1---- 1---- 2---- 2
|     |     |     |     |
0---- 0---- 1---- 1---- 1
|     |     |     |     |
0---- 0---- 0---- 0---- 0
|     |     |     |     |
0---- 0---- 0---- 0---- 0


We assume that after a load balancing step, we have the following
distribution:

1---- 1---- 1---- 3---- 3
|     |     |     |     |
1---- 1---- 1---- 3---- 3
|     |     |     |     |
0---- 0---- 0---- 3---- 3
|     |     |     |     |
0---- 0---- 2---- 2---- 2
|     |     |     |     |
0---- 0---- 2---- 2---- 2

We create a ParticleArray to represent the initial distribution and
use ParticleArrayExchange to move the data by manually setting the
particle import/export lists. We require that the test be run with 4
processors.

"""
import mpi4py.MPI as mpi
import numpy as np

from pysph.parallel.parallel_manager import ParticleArrayExchange
from pysph.base.utils import get_particle_array_wcsph


def main():
    comm = mpi.COMM_WORLD
    rank = comm.Get_rank()
    size = comm.Get_size()

    if size != 4:
        if rank == 0:
            raise RuntimeError("Run this test with 4 processors")

    # create the initial distribution
    if rank == 0:
        numPoints = 12
        x = np.array([0.0, 1.0, 2.0, 3.0, 4.0,
                      0.0, 1.0, 2.0, 3.0, 4.0,
                      0.0, 1.0])

        y = np.array([0.0, 0.0, 0.0, 0.0, 0.0,
                      1.0, 1.0, 1.0, 1.0, 1.0,
                      2.0, 2.0])

        gid = np.array([0, 1, 2, 3, 4, 5,
                        6, 7, 8, 9, 10,
                        11], dtype=np.uint32)

        exportLocalids = np.array([2, 3, 4, 7, 8, 9], dtype=np.uint32)
        exportProcs = np.array([2, 2, 2, 2, 2, 2], dtype=np.int32)
        numExport = 6

        parts = np.ones(shape=numPoints, dtype=np.int32) * rank

    if rank == 1:
        numPoints = 6
        x = np.array([2.0, 3.0, 4.0,
                      0.0, 1.0, 2.0])

        y = np.array([2.0, 2.0, 2.0,
                      3.0, 3.0, 3.0])

        gid = np.array([12, 13, 14, 15,
                        16, 17], dtype=np.uint32)

        exportLocalids = np.array([0, 1, 2], dtype=np.uint32)
        exportProcs = np.array([0, 3, 3], dtype=np.int32)
        numExport = 3

        parts = np.ones(shape=numPoints, dtype=np.int32) * rank

    if rank == 2:
        numPoints = 3
        x = np.array([4.0, 5.0, 0.0])
        y = np.array([3.0, 3.0, 4.0])
        gid = np.array([18, 19, 20], dtype=np.uint32)

        exportLocalids = np.array([0, 1, 2], dtype=np.uint32)
        exportProcs = np.array([3, 3, 1], dtype=np.int32)
        numExport = 3

        parts = np.ones(shape=numPoints, dtype=np.int32) * rank

    if rank == 3:
        numPoints = 4
        x = np.array([1.0, 2.0, 3.0, 4.0])
        y = np.array([4.0, 4.0, 4.0, 4.0])
        gid = np.array([21, 22, 23, 24], dtype=np.uint32)

        exportLocalids = np.array([0, 1], dtype=np.uint32)
        exportProcs = np.array([1, 1], dtype=np.int32)
        numExport = 2

        parts = np.ones(shape=numPoints, dtype=np.int32) * rank

    # Gather the Global data on root
    X = np.zeros(shape=25, dtype=np.float64)
    Y = np.zeros(shape=25, dtype=np.float64)
    GID = np.zeros(shape=25, dtype=np.uint32)

    displacements = np.array([12, 6, 3, 4], dtype=np.int32)

    comm.Gatherv(
        sendbuf=[x, mpi.DOUBLE], recvbuf=[X, (displacements, None)], root=0)
    comm.Gatherv(
        sendbuf=[y, mpi.DOUBLE], recvbuf=[Y, (displacements, None)], root=0)
    comm.Gatherv(
        sendbuf=[gid, mpi.UNSIGNED_INT],
        recvbuf=[GID, (displacements, None)],
        root=0)

    # broadcast global X, Y and GID to everyone
    comm.Bcast(buf=X, root=0)
    comm.Bcast(buf=Y, root=0)
    comm.Bcast(buf=GID, root=0)

    # create the local particle arrays and exchange objects
    u = np.ones_like(x)
    pa = get_particle_array_wcsph(name='test', x=x, y=y, u=u, gid=gid)
    # Squeeze the arrays so that they are forced to resize when doing
    # the lb exchange.
    for arr in pa.properties.values():
        arr.squeeze()
    pa.set_lb_props(['x', 'y', 'u', 'gid'])

    pae = ParticleArrayExchange(pa_index=0, pa=pa, comm=comm)

    # set the export indices for each array
    pae.reset_lists()
    pae.numParticleExport = numExport
    pae.exportParticleLocalids.resize(numExport)
    pae.exportParticleLocalids.set_data(exportLocalids)

    pae.exportParticleProcs.resize(numExport)
    pae.exportParticleProcs.set_data(exportProcs)

    # call lb_balance with these lists
    pae.lb_exchange_data()

    # All arrays must be local after lb_exchange_data
    assert (pa.num_real_particles == pa.get_number_of_particles())
    assert (np.allclose(pa.tag, 0))

    # now check the data on each array
    numParticles = 6
    if rank == 0:
        numParticles = 7

    assert (pa.num_real_particles == numParticles)

    for i in range(numParticles):
        assert (abs(X[pa.gid[i]] - pa.x[i]) < 1e-15)
        assert (abs(Y[pa.gid[i]] - pa.y[i]) < 1e-15)
        assert (GID[pa.gid[i]] == pa.gid[i])

    # Check that all non-load balanced props are not garbage values
    for prop in ('z', 'v', 'w', 'p'):
        np.testing.assert_array_almost_equal(pa.get(prop), 0.0)
    np.testing.assert_array_almost_equal(pa.u, 1.0)


if __name__ == '__main__':
    main()



"""A sphere of density 500 falling into a hydrostatic tank (15 minutes)

Check basic equations of SPH to throw a ball inside the vessel
"""
from __future__ import print_function
import numpy as np

# PySPH base and carray imports
from pysph.base.utils import (get_particle_array_wcsph,
                              get_particle_array_rigid_body)
from pysph.base.kernels import CubicSpline

from pysph.solver.solver import Solver
from pysph.sph.integrator import EPECIntegrator
from pysph.sph.integrator_step import WCSPHStep

from pysph.sph.equation import Group
from pysph.sph.basic_equations import (XSPHCorrection, SummationDensity)
from pysph.sph.wc.basic import TaitEOSHGCorrection, MomentumEquation
from pysph.solver.application import Application
from pysph.sph.rigid_body import (
    BodyForce, SummationDensityBoundary, RigidBodyCollision, RigidBodyMoments,
    RigidBodyMotion, AkinciRigidFluidCoupling, RK2StepRigidBody)


def create_boundary():
    dx = 2

    # bottom particles in tank
    xb = np.arange(-2 * dx, 100 + 2 * dx, dx)
    yb = np.arange(-2 * dx, 0, dx)
    xb, yb = np.meshgrid(xb, yb)
    xb = xb.ravel()
    yb = yb.ravel()

    xl = np.arange(-2 * dx, 0, dx)
    yl = np.arange(0, 250, dx)
    xl, yl = np.meshgrid(xl, yl)
    xl = xl.ravel()
    yl = yl.ravel()

    xr = np.arange(100, 100 + 2 * dx, dx)
    yr = np.arange(0, 250, dx)
    xr, yr = np.meshgrid(xr, yr)
    xr = xr.ravel()
    yr = yr.ravel()

    x = np.concatenate([xl, xb, xr])
    y = np.concatenate([yl, yb, yr])

    return x * 1e-3, y * 1e-3


def create_fluid():
    dx = 2
    xf = np.arange(0, 100, dx)
    yf = np.arange(0, 150, dx)
    xf, yf = np.meshgrid(xf, yf)
    xf = xf.ravel()
    yf = yf.ravel()

    return xf * 1e-3, yf * 1e-3


def create_sphere(dx=1):
    x = np.arange(0, 100, dx)
    y = np.arange(151, 251, dx)
    x, y = np.meshgrid(x, y)
    x = x.ravel()
    y = y.ravel()

    p = ((x - 50)**2 + (y - 200)**2) < 20**2
    x = x[p]
    y = y[p]

    # lower sphere a little
    y = y - 20
    return x * 1e-3, y * 1e-3


def get_density(y):
    height = 150
    c_0 = 2 * np.sqrt(2 * 9.81 * height * 1e-3)
    rho_0 = 1000
    height_water_clmn = height * 1e-3
    gamma = 7.
    _tmp = gamma / (rho_0 * c_0**2)

    rho = np.zeros_like(y)
    for i in range(len(rho)):
        p_i = rho_0 * 9.81 * (height_water_clmn - y[i])
        rho[i] = rho_0 * (1 + p_i * _tmp)**(1. / gamma)
    return rho


def geometry():
    import matplotlib.pyplot as plt
    # please run this function to know how
    # geometry looks like
    x_tank, y_tank = create_boundary()
    x_fluid, y_fluid = create_fluid()
    x_cube, y_cube = create_sphere()
    plt.scatter(x_fluid, y_fluid)
    plt.scatter(x_tank, y_tank)
    plt.scatter(x_cube, y_cube)
    plt.axes().set_aspect('equal', 'datalim')
    print("done")
    plt.show()


class RigidFluidCoupling(Application):
    def initialize(self):
        self.dx = 2 * 1e-3
        self.hdx = 1.2
        self.ro = 1000
        self.solid_rho = 500
        self.m = 1000 * self.dx * self.dx
        self.co = 2 * np.sqrt(2 * 9.81 * 150 * 1e-3)
        self.alpha = 0.1

    def create_particles(self):
        """Create the circular patch of fluid."""
        xf, yf = create_fluid()
        m = self.ro * self.dx * self.dx
        rho = self.ro
        h = self.hdx * self.dx
        fluid = get_particle_array_wcsph(x=xf, y=yf, h=h, m=m, rho=rho,
                                         name="fluid")

        dx = 2
        xt, yt = create_boundary()
        m = 1000 * self.dx * self.dx
        rho = 1000
        rad_s = 2 / 2. * 1e-3
        h = self.hdx * self.dx
        V = dx * dx * 1e-6
        tank = get_particle_array_wcsph(x=xt, y=yt, h=h, m=m, rho=rho,
                                        rad_s=rad_s, V=V, name="tank")
        for name in ['fx', 'fy', 'fz']:
            tank.add_property(name)

        dx = 1
        xc, yc = create_sphere(1)
        m = self.solid_rho * dx * 1e-3 * dx * 1e-3
        rho = self.solid_rho
        h = self.hdx * self.dx
        rad_s = dx / 2. * 1e-3
        V = dx * dx * 1e-6
        cs = 0.0
        cube = get_particle_array_rigid_body(x=xc, y=yc, h=h, m=m, rho=rho,
                                             rad_s=rad_s, V=V, cs=cs,
                                             name="cube")

        return [fluid, tank, cube]

    def create_solver(self):
        kernel = CubicSpline(dim=2)

        integrator = EPECIntegrator(fluid=WCSPHStep(), tank=WCSPHStep(),
                                    cube=RK2StepRigidBody())

        dt = 0.125 * self.dx * self.hdx / (self.co * 1.1) / 2.
        # dt = 1e-4
        print("DT: %s" % dt)
        tf = 0.5
        solver = Solver(
            kernel=kernel,
            dim=2,
            integrator=integrator,
            dt=dt,
            tf=tf,
            adaptive_timestep=False, )

        return solver

    def create_equations(self):
        equations = [
            Group(equations=[
                BodyForce(dest='cube', sources=None, gy=-9.81),
            ], real=False),
            Group(equations=[
                SummationDensity(
                    dest='fluid',
                    sources=['fluid'], ),
                SummationDensityBoundary(
                    dest='fluid', sources=['tank', 'cube'], fluid_rho=1000.0)
            ]),

            # Tait equation of state
            Group(equations=[
                TaitEOSHGCorrection(dest='fluid', sources=None, rho0=self.ro,
                                    c0=self.co, gamma=7.0),
            ], real=False),
            Group(equations=[
                MomentumEquation(dest='fluid', sources=['fluid'],
                                 alpha=self.alpha, beta=0.0, c0=self.co,
                                 gy=-9.81),
                AkinciRigidFluidCoupling(dest='fluid',
                                         sources=['cube', 'tank']),
                XSPHCorrection(dest='fluid', sources=['fluid', 'tank']),
            ]),
            Group(equations=[
                RigidBodyCollision(dest='cube', sources=['tank'], kn=1e5)
            ]),
            Group(equations=[RigidBodyMoments(dest='cube', sources=None)]),
            Group(equations=[RigidBodyMotion(dest='cube', sources=None)]),
        ]
        return equations


if __name__ == '__main__':
    app = RigidFluidCoupling()
    app.run()



from mako.template import Template
from mako.lookup import TemplateLookup
import os
import sys

from compyle.opencl import get_context, profile_kernel, SimpleKernel


def get_simple_kernel(kernel_name, args, src, wgs, preamble=""):
    ctx = get_context()
    knl = SimpleKernel(
        ctx, args, src, wgs,
        kernel_name, preamble=preamble
    )

    return profile_kernel(knl, kernel_name, backend='opencl')


def get_elwise_kernel(kernel_name, args, src, preamble=""):
    ctx = get_context()
    from pyopencl.elementwise import ElementwiseKernel
    knl = ElementwiseKernel(
        ctx, args, src,
        kernel_name, preamble=preamble
    )
    return profile_kernel(knl, kernel_name, backend='opencl')


class GPUNNPSHelper(object):
    def __init__(self, tpl_filename, backend=None, use_double=False,
                 c_type=None):
        """

        Parameters
        ----------
        tpl_filename
            filename of source template
        backend
            backend to use for helper
        use_double:
            Use double precision floating point data types
        c_type:
            c_type to use. Overrides use_double
        """
        disable_unicode = False if sys.version_info.major > 2 else True

        self.src_tpl = Template(
            filename=os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                tpl_filename),
            disable_unicode=disable_unicode,
        )
        self.data_t = "double" if use_double else "float"

        if c_type is not None:
            self.data_t = c_type

        helper_tpl = Template(
            filename=os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                "gpu_helper_functions.mako"),
            disable_unicode=disable_unicode
        )

        helper_preamble = helper_tpl.get_def("get_helpers").render(
            data_t=self.data_t
        )
        preamble = self.src_tpl.get_def("preamble").render(
            data_t=self.data_t
        )
        self.preamble = "\n".join([helper_preamble, preamble])
        self.cache = {}
        self.backend = backend

    def _get_code(self, kernel_name, **kwargs):
        arguments = self.src_tpl.get_def("%s_args" % kernel_name).render(
            data_t=self.data_t, **kwargs)

        src = self.src_tpl.get_def("%s_src" % kernel_name).render(
            data_t=self.data_t, **kwargs)

        return arguments, src

    def get_kernel(self, kernel_name, **kwargs):
        key = kernel_name, tuple(kwargs.items())
        wgs = kwargs.get('wgs', None)

        if key in self.cache:
            return self.cache[key]
        else:
            args, src = self._get_code(kernel_name, **kwargs)

            if wgs is None:
                knl = get_elwise_kernel(kernel_name, args, src,
                                        preamble=self.preamble)
            else:
                knl = get_simple_kernel(kernel_name, args, src, wgs,
                                        preamble=self.preamble)

            self.cache[key] = knl
            return knl



"""Utility module to read input mesh files.  This is primarily for meshes
generated using Gmsh.  This module also provides some simple classes
that allow one to create extruded 3D surfaces by generating a gmsh file
in Python.

There is also a function to read VTK dataset and produce points from them.
This is very useful as Gmsh can generate VTK datasets from its meshes and thus
the meshes can be imported as point clouds that may be used in an SPH
simulation.
"""
# Copyright (c) 2015 Prabhu Ramachandran

import gzip
import json

import numpy as np
import subprocess

import tempfile
from tvtk.api import tvtk

import os
from os.path import exists, expanduser, join
import sys

def _read_vtk_file(fname):
    """Given a .vtk file (or .vtk.gz), read it and return the output.
    """
    if fname.endswith('.vtk.gz'):
        tmpfname = tempfile.mktemp(suffix='.vtk')
        with open(tmpfname, 'wb') as tmpf:
            data = gzip.open(fname).read()
            tmpf.write(data)
        r = tvtk.DataSetReader(file_name=tmpfname)
    else:
        tmpfname = None
        r = tvtk.DataSetReader(file_name=fname)
    r.update()
    if tmpfname is not None:
        os.remove(tmpfname)
    return r.output

def _convert_to_points(dataset, vertices=True, cell_centers=True):
    """Given a VTK dataset, convert it to a set of points that can be used
    for simulation with SPH.

    Parameters
    ----------

    dataset : tvtk.DataSet
    vertices: bool
        If True, it converts the vertices to points.
    cell_centers: bool
        If True, converts the cell centers to points.

    Returns
    -------

    x, y, z of the points.
    """
    pts = np.array([], dtype=float)
    if vertices:
        pts = np.append(pts, dataset.points.to_array())
    if cell_centers:
        cell_centers = tvtk.CellCenters(input=dataset)
        cell_centers.update()
        p = cell_centers.output.points.to_array()
        pts = np.append(pts, p)
    pts.shape = len(pts)//3, 3
    x, y, z = pts.T
    return x, y, z

def vtk_file_to_points(fname, vertices=True, cell_centers=True):
    """Given a file containing a VTK dataset (currently only an old style .vtk
    file), convert it to a set of points that can be used for simulation with
    SPH.

    Parameters
    ----------

    fname : str
        File name.
    vertices: bool
        If True, it converts the vertices to points.
    cell_centers: bool
        If True, converts the cell centers to points.

    Returns
    -------

    x, y, z of the points.
    """
    dataset = _read_vtk_file(fname)
    return _convert_to_points(dataset, vertices, cell_centers)

def transform_points(x, y, z, transform):
    """Given the coordinates, x, y, z and the TVTK transform instance,
    return the transformed coordinates.
    """
    assert isinstance(transform, tvtk.Transform)

    m = transform.matrix.to_array()
    xt, yt, zt, wt = np.dot(m, np.array((x, y, z, np.ones_like(x))))
    return xt, yt, zt


class Loop(object):
    """Create a Line Loop in Gmsh parlance but using a turtle-graphics like
    approach.

    Use this to create a 2D closed surface.  The surface is always in the x-y
    plane.

    Examples
    --------

    Here is a simple example::

      >>> l1 = Loop((0.0, 0.0), mesh_size=0.1)
      >>> l1.move(1.0).turn(90).move(1.0).turn(90).move(1.0).turn(90).move(1.0)

    This will create a square shape.

    """

    def __init__(self, start, mesh_size=0.1):
        self.mesh_size = mesh_size
        self.points = [start]
        self.elems = []
        self.tolerance = 1e-12
        self._last_angle = 0.0
        self._last_position = start
        self._index = 1

    ### Public Protocol ###################################
    def turn(self, angle):
        """Turn by angle (in degrees).
        """
        self._last_angle += angle
        return self

    def move(self, dist):
        """Move by given distance at the current angle.
        """
        x0, y0 = self._last_position
        angle = self._last_angle
        rad = np.pi*angle/180.
        x = x0 + dist*np.cos(rad)
        y = y0 + dist*np.sin(rad)
        p1 = self._index
        p2 = self._add_point(x, y)
        self._add_elem('line', (p1, p2))
        return self

    def arc(self, radius, angle=180):
        """Create a circular arc given the radius and for an angle as
        specified.
        """
        x0, y0 = self._last_position
        last_angle = self._last_angle
        rad = np.pi*last_angle/180.
        last = complex(x0, y0)
        dz = complex(np.cos(rad), np.sin(rad))
        cen = last + radius*dz
        rad = np.pi*angle/180.
        end = cen + (last - cen)*complex(np.cos(rad), np.sin(rad))
        s_idx = self._index
        c_idx = self._add_point(cen.real, cen.imag)
        e_idx = self._add_point(end.real, end.imag)
        self._add_elem('circle', (s_idx, c_idx, e_idx))
        return self

    def write(self, fp, point_id_base=0, elem_id_base=0):
        """Write data to given file object `fp`.
        """
        points = self.points
        for i in range(len(points)):
            idx = point_id_base + i + 1
            self._write_point(fp, points[i], idx)

        elems = self.elems
        idx = 0
        loop = []
        for i in range(len(elems)):
            elem = elems[i]
            idx = elem_id_base + i + 1
            loop.append(str(idx))
            kind, data = elem[0], elem[1]
            if kind == 'line':
                self._write_line(fp, data, idx, point_id_base)
            elif kind == 'circle':
                self._write_circle(fp, data, idx, point_id_base)
        idx += 1
        ids = ', '.join(loop)
        s = 'Line Loop({idx}) = {{{ids}}};\n'.format(idx=idx, ids=ids)
        fp.write(s)
        return len(points), len(elems) + 1

    ### Private Protocol #################################
    def _add_elem(self, kind, data):
        self.elems.append((kind, data))

    def _check_for_existing_point(self, x, y):
        tol = self.tolerance
        for i, (xi, yi) in enumerate(self.points):
            if abs(xi - x) < tol and abs(yi - y) < tol:
                return i + 1

    def _add_point(self, x, y):
        last_x, last_y = self._last_position
        pnt = (x, y)
        existing = self._check_for_existing_point(x, y)
        if existing is None:
            self.points.append(pnt)
            self._index += 1
            index = self._index
        else:
            index = existing
            pnt = self.points[existing]
        self._last_position = pnt
        return index

    def _write_point(self, fp, pnt, idx):
        fp.write('Point({idx}) = {{{x}, {y}, 0.0, {mesh_size}}};\n'.\
            format(x=pnt[0], y=pnt[1], idx=idx, mesh_size=self.mesh_size))

    def _write_line(self, fp, data, e_idx, p_idx):
        s = 'Line({idx}) = {{{p1}, {p2}}};\n'.format(
            idx=e_idx, p1=data[0]+p_idx, p2=data[1] +p_idx
        )
        fp.write(s)

    def _write_circle(self, fp, data, e_idx, p_idx):
        s = 'Circle({idx}) = {{{start}, {cen}, {end}}};\n'.format(
            idx=e_idx, start=data[0]+p_idx, cen=data[1]+p_idx, end=data[2] +p_idx
        )
        fp.write(s)


class Surface(object):
    def __init__(self, *loops):
        """Constructor.

        Parameters
        ----------

        loops : tuple(Loop)
            Any additional positional arguments are treated as loop objects.
        """
        self.loops = list(loops)
        self.idx = 0

    def write(self, fp, point_id_base=0, elem_id_base=0):
        pid_base = point_id_base
        eid_base = elem_id_base
        loop_ids = []
        for loop in self.loops:
            np, ne = loop.write(fp, pid_base, eid_base)
            pid_base += np
            eid_base += ne
            loop_ids.append(str(eid_base))
        idx = eid_base + 1
        loop_str = ', '.join(loop_ids)
        fp.write('Plane Surface({idx}) = {{{loop_str}}};\n'.format(
                idx=idx, loop_str=loop_str))
        self.idx = idx
        return pid_base, eid_base


class Extrude(object):
    def __init__(self, dx=0.0, dy=0.0, dz=1.0, surfaces=None):
        """Extrude a given set of surfaces by the displacements given
        along each directions.

        Parameters
        ----------

        dx : float
            Extrusion along x.
        dy : float
            Extrusion along y.
        dz : float
            Extrusion along z.
        surfaces: list
            List of surfaces to extrude.

        """
        self.dx, self.dy, self.dz = dx, dy, dz
        if surfaces is None:
            self.surfaces = []
        else:
            self.surfaces = list(surfaces)

    def write(self, fp, point_id_base=0, elem_id_base=0):
        pid_base = point_id_base
        eid_base = elem_id_base
        surf_ids = []
        for surf in self.surfaces:
            np, ne = surf.write(fp, pid_base, eid_base)
            pid_base += np
            eid_base += ne
            surf_ids.append(str(surf.idx))

        s_ids = ', '.join(surf_ids)
        fp.write(
            'Extrude {{{dx}, {dy}, {dz}}} {{\n'
            '    Surface{{{s_ids}}};\n}}\n'.format(
            dx=self.dx, dy=self.dy, dz=self.dz, s_ids=s_ids
            )
        )
        return pid_base, eid_base


class Gmsh(object):
    def __init__(self, gmsh=None):
        """Construct a Gmsh helper object that can be used to mesh objects.

        Parameters
        ----------

        gmsh: str
            Path to gmsh executable.
        """
        self.config = expanduser(join('~', '.pysph', 'gmsh.json'))
        if gmsh is None:
            if exists(self.config):
                self._read_config()
            else:
                gmsh = self._ask_user_for_gmsh()
                self._set_gmsh(gmsh)
        else:
            self._set_gmsh(gmsh)

    #### Public Protocol ##################################
    def write_geo(self, entities, fp):
        """Write a list of given entities to the given file pointer.
        """
        p_count, e_count = 0, 0
        for entity in entities:
            np, ne = entity.write(fp, p_count, e_count)
            p_count += np
            e_count += ne

    def write_vtk_mesh(self, entities, fname):
        """Write a list of given entities to the given file name.
        """
        tmp_geo = tempfile.mktemp(suffix='.geo')
        try:
            self.write_geo(entities, open(tmp_geo, 'w'))
            self._call_gmsh('-3', tmp_geo, '-o', fname)
        finally:
            os.remove(tmp_geo)

    def get_points(self, entities, vertices=True, cell_centers=False):
        """Given a list of entities, return x, y, z arrays for the position.

        Parameters
        ----------

        entities : list
            List of entities.
        vertices: bool
            If True, it converts the vertices to points.
        cell_centers: bool
            If True, converts the cell centers to points.

        """
        tmp_vtk = tempfile.mktemp(suffix='.vtk')
        try:
            self.write_vtk_mesh(entities, tmp_vtk)
            x, y, z = vtk_file_to_points(
                tmp_vtk, vertices=vertices, cell_centers=cell_centers
            )
            return x, y, z
        finally:
            os.remove(tmp_vtk)

    def get_points_from_geo(self, geo_file_name, vertices=True,
                            cell_centers=False):
        """Given a .geo file, generate a mesh and get the points from the mesh.

        Parameters
        ----------

        geo_file_name: str
            Filename of the .geo file.
        vertices: bool
            If True, it converts the vertices to points.
        cell_centers: bool
            If True, converts the cell centers to points.

        """
        tmp_vtk = tempfile.mktemp(suffix='.vtk')
        try:
            self._call_gmsh('-3', geo_file_name, '-o', tmp_vtk)
            return vtk_file_to_points(
                tmp_vtk, vertices=vertices, cell_centers=cell_centers
            )
        finally:
            os.remove(tmp_vtk)

    #### Private Protocol #################################
    def _ask_user_for_gmsh(self):
        gmsh = input('Please provide the path to gmsh executable: ')
        return gmsh

    def _read_config(self):
        if exists(self.config):
            data = json.load(open(self.config))
            self.gmsh = data['path']

    def _set_gmsh(self, gmsh):
        self.gmsh = gmsh
        data = dict(path=gmsh)
        json.dump(data, open(self.config, 'w'))

    def _call_gmsh(self, *args):
        if self.gmsh is None:
            raise RuntimeError('Gmsh is not configured, set the gmsh path.')
        cmd = [self.gmsh] + list(args)
        subprocess.check_call(cmd)


def example_3d_p(fp=sys.stdout):
    """Creates a 3D "P" with a hole inside it.
    """
    # The exterior of the "P"
    l1 = Loop((0.0, 0.0), mesh_size=0.1)
    l1.turn(-90).move(1.0).turn(90).move(0.2).turn(90).move(0.5)\
       .arc(0.25, -180).turn(90).move(0.2)
    # The inner loop for the hole in the middle.
    l2 = Loop((0.1, -0.25))
    l2.arc(0.1, 90).turn(90).arc(0.1, 90).turn(90)\
      .arc(0.1, 90).turn(90).arc(0.1, 90)
    s = Surface(l1, l2)
    ex = Extrude(0.0, 0.0, 1.0, surfaces=[s])
    ex.write(fp)
    return ex

def example_cube(fp=sys.stdout):
    """Simple example of a cube.
    """
    l1 = Loop((0.0, 0.0), mesh_size=0.1)
    l1.move(1.0).turn(90).move(1.0).turn(90).move(1.0).turn(90).move(1.0)
    s = Surface(l1)
    ex = Extrude(0.0, 0.0, 1.0, surfaces=[s])
    ex.write(fp)
    return ex

def example_plot_3d_p(gmsh):
    """Note: this will only work if you have gmsh installed.
    """
    import io
    fp = io.StringIO()
    ex = example_3d_p(fp)
    g = Gmsh(gmsh)
    x, y, z = g.get_points([ex])
    from mayavi import mlab
    mlab.points3d(x, y, z, color=(1, 0, 0))



'''Tests for integrator having different acceleration evaluators for different
stages.

'''
import unittest

import pytest
import numpy as np

from pysph.base.utils import get_particle_array
from pysph.sph.equation import Equation, MultiStageEquations
from pysph.sph.acceleration_eval import make_acceleration_evals
from pysph.base.kernels import CubicSpline
from pysph.base.nnps import LinkedListNNPS as NNPS
from pysph.sph.sph_compiler import SPHCompiler
from pysph.sph.integrator import Integrator
from pysph.sph.integrator_step import IntegratorStep


class Eq1(Equation):
    def initialize(self, d_idx, d_au):
        d_au[d_idx] = 1.0


class Eq2(Equation):
    def initialize(self, d_idx, d_au):
        d_au[d_idx] += 1.0


class MyStepper(IntegratorStep):
    def stage1(self, d_idx, d_u, d_au, dt):
        d_u[d_idx] += d_au[d_idx]*dt

    def stage2(self, d_idx, d_u, d_au, dt):
        d_u[d_idx] += d_au[d_idx]*dt


class MyIntegrator(Integrator):
    def one_timestep(self, t, dt):
        self.compute_accelerations(0, update_nnps=False)
        self.stage1()
        self.do_post_stage(dt, 1)
        self.compute_accelerations(1, update_nnps=False)
        self.stage2()
        self.update_domain()
        self.do_post_stage(dt, 2)


class TestMultiGroupIntegrator(unittest.TestCase):
    def setUp(self):
        self.dim = 1
        n = 10
        dx = 1.0/(n-1)
        x = np.linspace(0, 1, n)
        m = np.ones_like(x)
        h = np.ones_like(x)*dx*1.05
        pa = get_particle_array(name='fluid', x=x, h=h, m=m, au=0.0, u=0.0)
        self.pa = pa
        self.NNPS_cls = NNPS
        self.backend = 'cython'

    def _make_integrator(self):
        arrays = [self.pa]
        kernel = CubicSpline(dim=self.dim)
        eqs = [
            [Eq1(dest='fluid', sources=['fluid'])],
            [Eq2(dest='fluid', sources=['fluid'])]
        ]
        meqs = MultiStageEquations(eqs)
        a_evals = make_acceleration_evals(
            arrays, meqs, kernel, backend=self.backend
        )
        integrator = MyIntegrator(fluid=MyStepper())
        comp = SPHCompiler(a_evals, integrator=integrator)
        comp.compile()
        nnps = self.NNPS_cls(dim=kernel.dim, particles=arrays)
        nnps.update()
        for ae in a_evals:
            ae.set_nnps(nnps)
        integrator.set_nnps(nnps)
        return integrator

    def test_different_accels_per_integrator(self):
        # Given
        pa = self.pa
        integrator = self._make_integrator()

        # When
        integrator.step(0.0, 0.1)

        # Then
        if pa.gpu:
            pa.gpu.pull('u', 'au')
        one = np.ones_like(pa.x)
        np.testing.assert_array_almost_equal(pa.au, 2.0*one)
        np.testing.assert_array_almost_equal(pa.u, 0.3*one)


class TestMultiGroupIntegratorGPU(TestMultiGroupIntegrator):
    def setUp(self):
        pytest.importorskip('pyopencl')
        pytest.importorskip('pysph.base.gpu_nnps')
        super(TestMultiGroupIntegratorGPU, self).setUp()
        from pysph.base.gpu_nnps import ZOrderGPUNNPS
        self.NNPS_cls = ZOrderGPUNNPS
        self.backend = 'opencl'


if __name__ == '__main__':
    unittest.main()



import unittest

import numpy as np

from pysph.base.utils import get_particle_array
from pysph.base.kernels import QuinticSpline

from pysph.sph.acceleration_eval import AccelerationEval
from pysph.sph.acceleration_eval_cython_helper import AccelerationEvalCythonHelper
from pysph.sph.integrator import PECIntegrator
from pysph.sph.integrator_step import WCSPHStep
from pysph.sph.integrator_cython_helper import IntegratorCythonHelper

from pysph.sph.basic_equations import SummationDensity


class TestIntegratorCythonHelper(unittest.TestCase):
    def test_invalid_kwarg_raises_error(self):
        # Given
        x = np.linspace(0, 1, 10)
        pa = get_particle_array(name='fluid', x=x)
        equations = [SummationDensity(dest='fluid', sources=['fluid'])]
        kernel = QuinticSpline(dim=1)
        a_eval = AccelerationEval([pa], equations, kernel=kernel)
        a_helper = AccelerationEvalCythonHelper(a_eval)

        # When/Then
        integrator = PECIntegrator(f=WCSPHStep())
        self.assertRaises(
            RuntimeError,
            IntegratorCythonHelper,
            integrator, a_helper
        )


if __name__ == '__main__':
    unittest.main()



#cython: embedsignature=True

# malloc and friends
from libc.stdlib cimport malloc, free
from libc.stdio cimport printf
from libcpp.vector cimport vector

from .nnps_base cimport *

# Cython for compiler directives
cimport cython

DEF EPS = 1e-6

IF UNAME_SYSNAME == "Windows":
    cdef inline double fmin(double x, double y) nogil:
        return x if x < y else y
    cdef inline double fmax(double x, double y) nogil:
        return x if x > y else y

#############################################################################
cdef class StratifiedHashNNPS(NNPS):

    """Finds nearest neighbors using Spatial Hashing with particles classified according
    to their support radii.
    """

    def __init__(self, int dim, list particles, double radius_scale = 2.0,
            int ghost_layers = 1, domain=None, bint fixed_h = False,
            bint cache = False, bint sort_gids = False, int H = 1,
            int num_levels = 1, long long int table_size = 131072):
        NNPS.__init__(
            self, dim, particles, radius_scale, ghost_layers, domain,
            cache, sort_gids
        )

        self.table_size = table_size
        self.radius_scale2 = radius_scale*radius_scale
        self.interval_size = 0
        self.H = H

        self.src_index = 0
        self.dst_index = 0
        self.sort_gids = sort_gids
        self.domain.update()
        self.update()

    def __cinit__(self, int dim, list particles, double radius_scale = 2.0,
            int ghost_layers = 1, domain=None, bint fixed_h = False,
            bint cache = False, bint sort_gids = False, int H = 1,
            int num_levels = 1, long long int table_size = 131072):

        cdef int narrays = len(particles)
        cdef HashTable** current_hash

        if fixed_h:
            self.num_levels = 1
        else:
            self.num_levels = num_levels

        self.hashtable = <HashTable***> malloc(narrays*sizeof(HashTable**))
        self.cell_sizes = <double**> malloc(narrays*sizeof(double*))

        cdef int i, j
        for i from 0<=i<narrays:
            self.hashtable[i] = <HashTable**> malloc(self.num_levels*sizeof(HashTable*))
            self.cell_sizes[i] = <double*> malloc(self.num_levels*sizeof(double))
            current_hash = self.hashtable[i]
            for j from 0<=j<self.num_levels:
                current_hash[j] = NULL

        self.current_hash = NULL
        self.current_cells = NULL

    def __dealloc__(self):
        cdef HashTable** current_hash
        cdef int i, j
        for i from 0<=i<self.narrays:
            current_hash = self.hashtable[i]
            for j from 0<=j<self.num_levels:
                if current_hash[j] != NULL:
                    del current_hash[j]
            free(self.hashtable[i])
            free(self.cell_sizes[i])
        free(self.hashtable)
        free(self.cell_sizes)

    #### Public protocol ################################################

    cpdef int count_particles(self, int interval):
        """Count number of particles in at a level"""
        cdef int i
        cdef int num_particles = 0
        return self.current_hash[interval].number_of_particles()

    cpdef double get_binning_size(self, int interval):
        """Get bin size at a level"""
        return self._get_h_max(self.current_cells, interval)

    cpdef set_context(self, int src_index, int dst_index):
        """Set context for nearest neighbor searches.

        Parameters
        ----------
        src_index: int
            Index in the list of particle arrays to which the neighbors belong

        dst_index: int
            Index in the list of particle arrays to which the query point belongs

        """
        NNPS.set_context(self, src_index, dst_index)
        self.current_hash = self.hashtable[src_index]
        self.current_cells = self.cell_sizes[src_index]

        self.dst = <NNPSParticleArrayWrapper> self.pa_wrappers[dst_index]
        self.src = <NNPSParticleArrayWrapper> self.pa_wrappers[src_index]

    @cython.cdivision(True)
    cdef void find_nearest_neighbors(self, size_t d_idx, UIntArray nbrs) nogil:
        """Low level, high-performance non-gil method to find neighbors.
        This requires that `set_context()` be called beforehand.  This method
        does not reset the neighbors array before it appends the
        neighbors to it.

        """
        cdef double* dst_x_ptr = self.dst.x.data
        cdef double* dst_y_ptr = self.dst.y.data
        cdef double* dst_z_ptr = self.dst.z.data
        cdef double* dst_h_ptr = self.dst.h.data

        cdef double* src_x_ptr = self.src.x.data
        cdef double* src_y_ptr = self.src.y.data
        cdef double* src_z_ptr = self.src.z.data
        cdef double* src_h_ptr = self.src.h.data

        cdef double x = dst_x_ptr[d_idx]
        cdef double y = dst_y_ptr[d_idx]
        cdef double z = dst_z_ptr[d_idx]
        cdef double h = dst_h_ptr[d_idx]

        cdef unsigned int* s_gid = self.src.gid.data
        cdef int orig_length = nbrs.length

        cdef int c_x, c_y, c_z
        cdef double* xmin = self.xmin.data
        cdef unsigned int i, j, k, n
        cdef vector[unsigned int] *candidates
        cdef int candidate_size = 0

        cdef double xij2 = 0
        cdef double hi2 = self.radius_scale2*h*h
        cdef double hj2 = 0

        cdef double h_max
        cdef int H, mask_len, num_boxes

        cdef int* x_boxes
        cdef int* y_boxes
        cdef int* z_boxes

        cdef HashTable* hash_level = NULL
        cdef HashEntry* candidate_cell = NULL

        cdef double hmax_level

        for i from 0<=i<self.num_levels:
            hash_level = self.current_hash[i]
            if hash_level.number_of_particles() == 0:
                continue

            hmax_level = self._get_h_max(self.current_cells, i)
            h_max = fmax(self.radius_scale*h, hmax_level)
            H = <int> ceil(h_max*self.H / hmax_level)

            mask_len = (2*H+1)*(2*H+1)*(2*H+1)

            x_boxes = <int*> malloc(mask_len*sizeof(int))
            y_boxes = <int*> malloc(mask_len*sizeof(int))
            z_boxes = <int*> malloc(mask_len*sizeof(int))

            find_cell_id_raw(
                    x - xmin[0],
                    y - xmin[1],
                    z - xmin[2],
                    hmax_level / self.H,
                    &c_x, &c_y, &c_z
                    )

            num_boxes = self._neighbor_boxes(c_x, c_y, c_z,
                    x_boxes, y_boxes, z_boxes, H)

            for j from 0<=j<num_boxes:
                candidate_cell = hash_level.get(x_boxes[j], y_boxes[j], z_boxes[j])
                if candidate_cell == NULL:
                    continue
                candidates = candidate_cell.get_indices()
                candidate_size = candidates.size()
                for k from 0<=k<candidate_size:
                    n = (candidates[0])[k]
                    hj2 = self.radius_scale2*src_h_ptr[n]*src_h_ptr[n]
                    xij2 = norm2(
                            src_x_ptr[n] - x,
                            src_y_ptr[n] - y,
                            src_z_ptr[n] - z
                            )
                    if (xij2 < hi2) or (xij2 < hj2):
                        nbrs.c_append(n)

            free(x_boxes)
            free(y_boxes)
            free(z_boxes)

        if self.sort_gids:
            self._sort_neighbors(
                &nbrs.data[orig_length], nbrs.length - orig_length, s_gid
            )


    #### Private protocol ################################################

    @cython.cdivision(True)
    cdef inline int _get_hash_id(self, double h) nogil:
        return <int> floor((self.radius_scale*h - self.hmin)/self.interval_size)

    cdef inline double _get_h_max(self, double* current_cells, int hash_id) nogil:
        return self.radius_scale*current_cells[hash_id]

    @cython.cdivision(True)
    cdef inline int _h_mask_exact(self, int* x, int* y, int* z,
            int H) nogil:
        cdef int length = 0
        cdef int s, t, u

        for s from -H<=s<=H:
            for t from -H<=t<=H:
                for u from -H<=u<=H:
                    x[length] = s
                    y[length] = t
                    z[length] = u
                    length += 1

        return length

    cdef inline int _neighbor_boxes(self, int i, int j, int k,
            int* x, int* y, int* z, int H) nogil:
        cdef int length = 0
        cdef int p

        cdef int mask_len = (2*H+1)*(2*H+1)*(2*H+1)

        cdef int* x_mask = <int*> malloc(mask_len*sizeof(int))
        cdef int* y_mask = <int*> malloc(mask_len*sizeof(int))
        cdef int* z_mask = <int*> malloc(mask_len*sizeof(int))

        mask_len = self._h_mask_exact(x_mask, y_mask, z_mask, H)

        for p from 0<=p<mask_len:
            if (i + x_mask[p] >= 0 and
                j + y_mask[p] >= 0 and
                k + z_mask[p] >= 0):
                    x[length] = i + x_mask[p]
                    y[length] = j + y_mask[p]
                    z[length] = k + z_mask[p]
                    length += 1

        free(x_mask)
        free(y_mask)
        free(z_mask)

        return length

    cdef inline void _set_h_max(self, double* current_cells, double* src_h_ptr,
            int num_particles) nogil:
        cdef double h
        cdef int i, idx
        for i from 0<=i<num_particles:
            h = src_h_ptr[i]
            idx = self._get_hash_id(h)
            current_cells[idx] = fmax(h, current_cells[idx])

    @cython.cdivision(True)
    cpdef _refresh(self):
        self.interval_size = (self.cell_size - self.hmin)/self.num_levels + EPS
        cdef HashTable** current_hash
        cdef int i, j
        for i from 0<=i<self.narrays:
            current_hash = self.hashtable[i]
            current_cells = self.cell_sizes[i]
            for j from 0<=j<self.num_levels:
                if current_hash[j] != NULL:
                    del current_hash[j]
                current_hash[j] = new HashTable(self.table_size)
                current_cells[j] = 0
        self.current_hash = self.hashtable[self.src_index]
        self.current_cells = self.cell_sizes[self.src_index]

    @cython.cdivision(True)
    cpdef _bin(self, int pa_index, UIntArray indices):
        cdef NNPSParticleArrayWrapper pa_wrapper = self.pa_wrappers[pa_index]

        cdef double* src_x_ptr = pa_wrapper.x.data
        cdef double* src_y_ptr = pa_wrapper.y.data
        cdef double* src_z_ptr = pa_wrapper.z.data
        cdef double* src_h_ptr = pa_wrapper.h.data

        cdef double* xmin = self.xmin.data

        cdef u_int i, idx
        cdef int hash_id
        cdef int c_x, c_y, c_z
        cdef double cell_size
        cdef double* current_cells = self.cell_sizes[pa_index]

        self._set_h_max(current_cells, src_h_ptr, pa_wrapper.get_number_of_particles())
        cdef HashTable** current_hash = self.hashtable[pa_index]

        for i from 0<=i<indices.length:
            idx = indices.data[i]
            hash_id = self._get_hash_id(src_h_ptr[idx])
            cell_size = self._get_h_max(current_cells, hash_id)/self.H
            find_cell_id_raw(
                    src_x_ptr[idx] - xmin[0],
                    src_y_ptr[idx] - xmin[1],
                    src_z_ptr[idx] - xmin[2],
                    cell_size,
                    &c_x, &c_y, &c_z
                    )
            current_hash[hash_id].add(c_x, c_y, c_z, idx, src_h_ptr[idx])



from math import sqrt
from compyle.api import declare

from pysph.sph.equation import Equation


class ComputeNormals(Equation):
    """Compute normals using a simple approach

    .. math::

       -\frac{m_j}{\rho_j} DW_{ij}

    First compute the normals, then average them and finally normalize them.

    """

    def initialize(self, d_idx, d_normal_tmp, d_normal):
        idx = declare('int')
        idx = 3*d_idx
        d_normal_tmp[idx] = 0.0
        d_normal_tmp[idx + 1] = 0.0
        d_normal_tmp[idx + 2] = 0.0
        d_normal[idx] = 0.0
        d_normal[idx + 1] = 0.0
        d_normal[idx + 2] = 0.0

    def loop(self, d_idx, d_normal_tmp, s_idx, s_m, s_rho, DWIJ):
        idx = declare('int')
        idx = 3*d_idx
        fac = -s_m[s_idx]/s_rho[s_idx]
        d_normal_tmp[idx] += fac*DWIJ[0]
        d_normal_tmp[idx + 1] += fac*DWIJ[1]
        d_normal_tmp[idx + 2] += fac*DWIJ[2]

    def post_loop(self, d_idx, d_normal_tmp, d_h):
        idx = declare('int')
        idx = 3*d_idx
        mag = sqrt(d_normal_tmp[idx]**2 + d_normal_tmp[idx + 1]**2 +
                   d_normal_tmp[idx + 2]**2)
        if mag > 0.25/d_h[d_idx]:
            d_normal_tmp[idx] /= mag
            d_normal_tmp[idx + 1] /= mag
            d_normal_tmp[idx + 2] /= mag
        else:
            d_normal_tmp[idx] = 0.0
            d_normal_tmp[idx + 1] = 0.0
            d_normal_tmp[idx + 2] = 0.0


class SmoothNormals(Equation):
    def loop(self, d_idx, d_normal, s_normal_tmp, s_idx, s_m, s_rho, WIJ):
        idx = declare('int')
        idx = 3*d_idx
        fac = s_m[s_idx]/s_rho[s_idx]*WIJ
        d_normal[idx] += fac*s_normal_tmp[3*s_idx]
        d_normal[idx + 1] += fac*s_normal_tmp[3*s_idx + 1]
        d_normal[idx + 2] += fac*s_normal_tmp[3*s_idx + 2]

    def post_loop(self, d_idx, d_normal, d_h):
        idx = declare('int')
        idx = 3*d_idx
        mag = sqrt(d_normal[idx]**2 + d_normal[idx + 1]**2 +
                   d_normal[idx + 2]**2)
        if mag > 1e-3:
            d_normal[idx] /= mag
            d_normal[idx + 1] /= mag
            d_normal[idx + 2] /= mag
        else:
            d_normal[idx] = 0.0
            d_normal[idx + 1] = 0.0
            d_normal[idx + 2] = 0.0


class SetWallVelocityNew(Equation):
    r"""Modified SetWall velocity which sets a suitable normal velocity.

    This requires that the destination array has a 3-strided "normal"
    property.
    """
    def initialize(self, d_idx, d_uf, d_vf, d_wf, d_wij):
        d_uf[d_idx] = 0.0
        d_vf[d_idx] = 0.0
        d_wf[d_idx] = 0.0
        d_wij[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_uf, d_vf, d_wf,
             s_u, s_v, s_w, d_wij, XIJ, RIJ, HIJ, SPH_KERNEL):
        wij = SPH_KERNEL.kernel(XIJ, RIJ, 0.5*HIJ)

        d_wij[d_idx] += wij
        d_uf[d_idx] += s_u[s_idx] * wij
        d_vf[d_idx] += s_v[s_idx] * wij
        d_wf[d_idx] += s_w[s_idx] * wij

    def post_loop(self, d_uf, d_vf, d_wf, d_wij, d_idx,
                  d_ug, d_vg, d_wg, d_u, d_v, d_w, d_normal):
        idx = declare('int')
        idx = 3*d_idx
        # calculation is done only for the relevant boundary particles.
        # d_wij (and d_uf) is 0 for particles sufficiently away from the
        # solid-fluid interface
        if d_wij[d_idx] > 1e-12:
            d_uf[d_idx] /= d_wij[d_idx]
            d_vf[d_idx] /= d_wij[d_idx]
            d_wf[d_idx] /= d_wij[d_idx]

        # Dummy velocities at the ghost points using Eq. (23),
        # d_u, d_v, d_w are the prescribed wall velocities.
        d_ug[d_idx] = 2*d_u[d_idx] - d_uf[d_idx]
        d_vg[d_idx] = 2*d_v[d_idx] - d_vf[d_idx]
        d_wg[d_idx] = 2*d_w[d_idx] - d_wf[d_idx]

        vn = (d_ug[d_idx]*d_normal[idx] + d_vg[d_idx]*d_normal[idx+1]
              + d_wg[d_idx]*d_normal[idx+2])
        if vn < 0:
            d_ug[d_idx] -= vn*d_normal[idx]
            d_vg[d_idx] -= vn*d_normal[idx+1]
            d_wg[d_idx] -= vn*d_normal[idx+2]



r"""Cheng and Shu's 1d acoustic wave propagation in 1d (1 min)

particles have properties according
to the following distribuion
.. math::
        \rho = \rho_0 + \Delta\rho sin(kx)
        p = 1.0
        u = 1 + 0.1sin(kx)

with :math:`\Delta\rho = 1` and :math:`k = 2\pi/\lambda`
where \lambda is the domain length.
.. math::
        \rho_0 = 2, \gamma = 1.4 and p_0 = 1.0
"""


# standard library and numpy imports
import numpy

# pysph imports
from pysph.base.utils import get_particle_array as gpa
from pysph.base.nnps import DomainManager
from pysph.solver.application import Application
from pysph.sph.scheme import GSPHScheme, SchemeChooser
from pysph.sph.gas_dynamics.psph import PSPHScheme
from pysph.sph.gas_dynamics.tsph import TSPHScheme


class ChengShu(Application):
    def initialize(self):
        self.xmin = 0.
        self.xmax = 1.
        self.gamma = 1.4
        self.p_0 = 1.
        self.c_0 = 1.
        self.delta_rho = 1
        self.n_particles = 1000
        self.domain_length = self.xmax - self.xmin
        self.dx = self.domain_length / (self.n_particles - 1)
        self.k = 2 * numpy.pi / self.domain_length
        self.hdx = 2.
        self.dt = 1e-4
        self.tf = 1.0
        self.dim = 1

    def create_domain(self):
        return DomainManager(
            xmin=self.xmin, xmax=self.xmax, periodic_in_x=True
        )

    def create_particles(self):
        x = numpy.linspace(
            self.xmin, self.xmax, self.n_particles
            )
        rho = 2 + numpy.sin(2 * numpy.pi * x)*self.delta_rho

        p = numpy.ones_like(x)

        u = 1 + 0.1 * numpy.sin(2 * numpy.pi * x)

        cs = numpy.sqrt(
            self.gamma * p / rho
        )
        h = numpy.ones_like(x) * self.dx * self.hdx
        m = numpy.ones_like(x) * self.dx * rho
        e = p / ((self.gamma - 1) * rho)

        fluid = gpa(
            name='fluid', x=x, p=p, rho=rho, u=u, h=h, m=m, e=e, cs=cs
        )

        self.scheme.setup_properties([fluid])

        return [fluid, ]

    def create_scheme(self):
        gsph = GSPHScheme(
            fluids=['fluid'], solids=[], dim=self.dim,
            gamma=self.gamma, kernel_factor=1.,
            g1=0., g2=0., rsolver=3, interpolation=1, monotonicity=1,
            interface_zero=True, hybrid=False, blend_alpha=5.0,
            niter=200, tol=1e-6
        )

        psph = PSPHScheme(
            fluids=['fluid'], solids=[], dim=self.dim, gamma=self.gamma,
            hfact=1.2
        )

        tsph = TSPHScheme(
            fluids=['fluid'], solids=[], dim=self.dim, gamma=self.gamma,
            hfact=1.2)

        s = SchemeChooser(
            default='gsph', gsph=gsph, psph=psph, tsph=tsph
        )

        return s

    def configure_scheme(self):
        s = self.scheme
        if self.options.scheme == 'gsph':
            s.configure_solver(
                dt=self.dt, tf=self.tf,
                adaptive_timestep=False, pfreq=1000
            )
        elif self.options.scheme in ['tsph', 'psph']:
            s.configure(hfact=1.2)
            s.configure_solver(dt=self.dt, tf=self.tf,
                               adaptive_timestep=False, pfreq=1000)


if __name__ == "__main__":
    app = ChengShu()
    app.run()



"""
Predictive-Corrective Incompressible SPH (PCISPH)
#################################################

References
-----------

    .. [SolPaj2009] B. Solenthaler, R. Pajarola "Predictive-Corrective
        Incompressible SPH", ACM Trans. Graph 28 (2009), pp. 1--6.

"""


import numpy as np
from pysph.sph.integrator import Integrator
from pysph.sph.equation import Equation, Group
from pysph.base.utils import get_particle_array
from pysph.sph.integrator_step import IntegratorStep
from pysph.sph.scheme import Scheme, add_bool_argument


def get_particle_array_pcisph(constants=None, **props):
    pcisph_props = [
        'au', 'av', 'aw', 'arho', 'dwij2', 'u0', 'v0', 'w0', 'aup', 'avp',
        'awp', 'x0', 'y0', 'z0', 'rho0'
    ]
    pa = get_particle_array(
        constants=constants, additional_props=pcisph_props, **props
    )
    pa.add_constant('iters', np.zeros(10000))
    pa.add_property('dw', stride=3)
    pa.add_output_arrays(['p', 'dwij2'])
    return pa


class PCISPHIntegrator(Integrator):
    def one_timestep(self, t, dt):
        self.initialize()

        self.compute_accelerations(0)

        self.stage1()
        self.update_domain()

        self.do_post_stage(dt, 1)

    def initial_acceleration(self, t, dt):
        pass


class PCISPHStep(IntegratorStep):
    def __init__(self, show_itercount=False):
        self.show_itercount = show_itercount
        self.index = 0

    def initialize(self, d_idx, d_u, d_v, d_w, d_u0, d_v0, d_w0,
                   d_x, d_y, d_z, d_x0, d_y0, d_z0, d_rho, d_rho0):
        d_u0[d_idx] = d_u[d_idx]
        d_v0[d_idx] = d_v[d_idx]
        d_w0[d_idx] = d_w[d_idx]

        d_x0[d_idx] = d_x[d_idx]
        d_y0[d_idx] = d_y[d_idx]
        d_z0[d_idx] = d_z[d_idx]

        d_rho0[d_idx] = d_rho[d_idx]

    def py_stage1(self, dst, t, dt):
        if self.show_itercount:
            print("Iteration count = ", dst.iters[self.index])
            self.index += 1

    def stage1(self, d_idx, d_u, d_v, d_w, d_au, d_av, d_aw, d_x, d_y, d_z,
               d_aup, d_avp, d_awp, d_u0, d_v0, d_w0, d_x0, d_y0, d_z0, dt):
        d_u[d_idx] = d_u0[d_idx] + dt * (d_au[d_idx] + d_aup[d_idx])
        d_v[d_idx] = d_v0[d_idx] + dt * (d_av[d_idx] + d_avp[d_idx])
        d_w[d_idx] = d_w0[d_idx] + dt * (d_aw[d_idx] + d_awp[d_idx])

        d_x[d_idx] = d_x0[d_idx] + dt * d_u[d_idx]
        d_y[d_idx] = d_y0[d_idx] + dt * d_v[d_idx]
        d_z[d_idx] = d_z0[d_idx] + dt * d_w[d_idx]


class MomentumEquationViscosity(Equation):
    r"""**Momentum Equation Viscosity**

    See "pysph.sph.wc.viscosity.LaminarViscocity"
    """

    def __init__(self, dest, sources, nu=0.0, gx=0.0, gy=0.0, gz=0.0):
        self.nu = nu
        self.gx = gx
        self.gy = gy
        self.gz = gz
        super(MomentumEquationViscosity, self).__init__(dest, sources)

    def initialize(self, d_idx, d_au, d_av, d_aw):
        d_au[d_idx] = self.gx
        d_av[d_idx] = self.gy
        d_aw[d_idx] = self.gz

    def loop(self, d_idx, s_idx, s_m, d_rho, s_rho, d_au, d_av, d_aw, DWIJ,
             XIJ, VIJ, R2IJ, EPS):
        mb = s_m[s_idx]
        rhoij = (d_rho[d_idx] + s_rho[s_idx])

        xdotdwij = DWIJ[0] * XIJ[0] + DWIJ[1] * XIJ[1] + DWIJ[2] * XIJ[2]

        tmp = mb * 4 * self.nu * xdotdwij / (rhoij * (R2IJ + EPS))

        d_au[d_idx] += tmp * VIJ[0]
        d_av[d_idx] += tmp * VIJ[1]
        d_aw[d_idx] += tmp * VIJ[2]

    def post_loop(self, d_idx, d_au, d_av, d_aw, d_u, d_v, d_w, d_p, d_aup,
                  d_avp, d_awp, dt):
        d_u[d_idx] += dt * d_au[d_idx]
        d_v[d_idx] += dt * d_av[d_idx]
        d_w[d_idx] += dt * d_aw[d_idx]

        # Retaining the old pressure seems to give better results for the
        # TG problem.
        #d_p[d_idx] = 0.0

        d_aup[d_idx] = 0.0
        d_avp[d_idx] = 0.0
        d_awp[d_idx] = 0.0


class Predict(Equation):
    r"""**Predict velocity and position**

    .. math::
        \mathbf{v}^{*}(t+1) = \mathbf{v}(t) + dt
        \left(\frac{d \mathbf{v}_{visc, g}(t)}{dt} +
        \frac{d \mathbf{v}_{p} (t)}{dt} \right)

    .. math::
        \mathbf{x}^{*}(t+1) = \mathbf{x}(t) + dt * \mathbf{v}(t+1)
    """

    def initialize(self, d_idx, d_u, d_v, d_w, d_aup, d_avp, d_awp, d_x, d_y,
                   d_z, d_au, d_av, d_aw, d_u0, d_v0, d_w0, d_x0, d_y0, d_z0,
                   dt):
        d_u[d_idx] = d_u0[d_idx] + dt * (d_au[d_idx] + d_aup[d_idx])
        d_v[d_idx] = d_v0[d_idx] + dt * (d_av[d_idx] + d_avp[d_idx])
        d_w[d_idx] = d_w0[d_idx] + dt * (d_aw[d_idx] + d_awp[d_idx])

        d_x[d_idx] = d_x0[d_idx] + dt * d_u[d_idx]
        d_y[d_idx] = d_y0[d_idx] + dt * d_v[d_idx]
        d_z[d_idx] = d_z0[d_idx] + dt * d_w[d_idx]


class ComputePressure(Equation):
    r"""**Compute Pressure**

    Compute pressure iteratively maintaining density within a given tolerance.

    .. math::
        p_i += \delta \rho^{*}_{{err}_i}

    where,

    .. math::
        \rho_{err_i} = \rho_i^{*} - \rho_0

    .. math::
        \delta = \frac{-1}{\beta (-\sum_j \nabla W_{ij} \cdot
        \sum_j \nabla W_{ij} - \sum_j \nabla W_{ij} \nabla W_{ij})}
    """

    def __init__(self, dest, sources, rho0):
        self.rho0 = rho0
        super(ComputePressure, self).__init__(dest, sources)

    def initialize(self, d_idx, d_dw, d_dwij2):
        d_dw[d_idx * 3 + 0] = 0.0
        d_dw[d_idx * 3 + 1] = 0.0
        d_dw[d_idx * 3 + 2] = 0.0

        d_dwij2[d_idx] = 0.0

    def loop(self, d_idx, d_dw, d_dwij2, DWIJ):
        d_dw[d_idx * 3 + 0] += DWIJ[0]
        d_dw[d_idx * 3 + 1] += DWIJ[1]
        d_dw[d_idx * 3 + 2] += DWIJ[2]

        dwij2 = DWIJ[0] * DWIJ[0] + DWIJ[1] * DWIJ[1] + DWIJ[2] * DWIJ[2]
        d_dwij2[d_idx] += dwij2

    def post_loop(self, d_idx, d_dw, d_m, dt, d_dwij2, d_p, d_rho):
        dwx = d_dw[d_idx * 3 + 0]
        dwy = d_dw[d_idx * 3 + 1]
        dwz = d_dw[d_idx * 3 + 2]
        tmp = dwx * dwx + dwy * dwy + dwz * dwz

        mi = d_m[d_idx]
        rho0 = self.rho0
        beta = 2 * mi * mi * (dt / rho0) * (dt / rho0)
        delta = 1.0 / (beta * (tmp + d_dwij2[d_idx]))

        rho_err = d_rho[d_idx] - rho0
        d_p[d_idx] += delta * rho_err


class MomentumEquationPressureGradient(Equation):
    r"""**Momentum Equation pressure gradient**

    Standard WCSPH pressure gradient,

    .. math::
        \frac{d\mathbf{v}}{dt} = - \sum_j m_j \left(\frac{p_i}{\rho_i^2}
        + \frac{p_i}{\rho_i^2}\right) \nabla W(x_{ij}, h)
    """

    def __init__(self, dest, sources, rho0, tolerance, debug):
        self.rho0 = rho0
        self.tolerance = tolerance
        self.debug = debug
        self.rho_err = 0.0
        self.ctr = 0
        super(MomentumEquationPressureGradient, self).__init__(dest, sources)

    def loop(self, d_idx, s_idx, d_p, s_p, d_rho, s_rho, s_m, d_aup, d_avp,
             d_awp, DWIJ):
        rhoi2 = 1.0 / (d_rho[d_idx] * d_rho[d_idx])
        rhoj2 = 1.0 / (s_rho[s_idx] * s_rho[s_idx])
        mj = s_m[d_idx]

        pij = -1.0 * mj * (d_p[d_idx] * rhoi2 + s_p[s_idx] * rhoj2)
        d_aup[d_idx] += pij * DWIJ[0]
        d_avp[d_idx] += pij * DWIJ[1]
        d_awp[d_idx] += pij * DWIJ[2]

    def reduce(self, dst, t, dt):
        import numpy as np
        self.rho_err = np.mean(np.abs(dst.rho / self.rho0 - 1.0))
        dst.iters[self.ctr] += 1

    def converged(self):
        debug = self.debug
        rho_err = self.rho_err

        if rho_err > self.tolerance:
            if debug:
                print("Not converged:", rho_err)
            return -1.0
        else:
            self.ctr += 1
            if debug:
                print("Converged:", rho_err)
            return 1.0


class PCISPHScheme(Scheme):
    def __init__(self, fluids, dim, rho0, nu, gx=0.0, gy=0.0, gz=0.0,
                 tolerance=0.1, debug=False, show_itercount=False):
        self.fluids = fluids
        self.solver = None
        self.dim = dim
        self.rho0 = rho0
        self.nu = nu
        self.gx = gx
        self.gy = gy
        self.gz = gz
        self.tolerance = tolerance
        self.debug = debug
        self.show_itercount = show_itercount

    def add_user_options(self, group):
        group.add_argument(
            '--pcisph-tol', action='store', type=float, dest='tolerance',
            default=None,
            help='relative error tolerance for convergence as a percentage.'
        )
        add_bool_argument(
            group, 'pcisph-debug', dest='debug', default=None,
            help="Produce some debugging output on convergence of iterations."
        )
        add_bool_argument(
            group, 'pcisph-itercount', dest='show_itercount', default=False,
            help="Produce some debugging output on convergence of iterations."
        )

    def consume_user_options(self, options):
        vars = ['tolerance', 'debug', 'show_itercount']
        data = dict((var, self._smart_getattr(options, var))
                    for var in vars)
        self.configure(**data)

    def configure_solver(self, kernel=None, integrator_cls=None,
                         extra_steppers=None, **kw):
        from pysph.base.kernels import QuinticSpline
        if kernel is None:
            kernel = QuinticSpline(dim=self.dim)
        steppers = {}
        if extra_steppers is not None:
            steppers.update(extra_steppers)

        step_cls = PCISPHStep
        for fluid in self.fluids:
            if fluid not in steppers:
                steppers[fluid] = step_cls(self.show_itercount)

        cls = PCISPHIntegrator if integrator_cls is None else integrator_cls
        integrator = cls(**steppers)

        from pysph.solver.solver import Solver
        self.solver = Solver(
            dim=self.dim, integrator=integrator, kernel=kernel,
            **kw
        )

    def get_equations(self):
        from pysph.sph.basic_equations import SummationDensity
        all = self.fluids
        equations = []

        eq1 = []
        for fluid in self.fluids:
            eq1.append(
                MomentumEquationViscosity(
                    dest=fluid, sources=all, nu=self.nu, gx=self.gx,
                    gy=self.gy, gz=self.gz
                )
            )
        equations.append(Group(equations=eq1))

        eq1, g2 = [], []
        for fluid in self.fluids:
            eq1.append(Predict(dest=fluid, sources=None))
        g2.append(Group(equations=eq1, update_nnps=True))

        eq2 = []
        for fluid in self.fluids:
            eq2.append(SummationDensity(dest=fluid, sources=all))
        g2.append(Group(equations=eq2))

        eq3 = []
        for fluid in self.fluids:
            eq3.append(
                ComputePressure(dest=fluid, sources=all, rho0=self.rho0)
            )
        g2.append(Group(equations=eq3, update_nnps=True))

        eq4 = []
        for fluid in self.fluids:
            eq4.append(
                MomentumEquationPressureGradient(
                    dest=fluid, sources=all, rho0=self.rho0,
                    tolerance=self.tolerance, debug=self.debug
                ),
            )
        g2.append(Group(equations=eq4))

        equations.append(
            Group(equations=g2, iterate=True,
                  max_iterations=500, min_iterations=2)
        )
        return equations

    def setup_properties(self, particles, clean=True):
        particle_arrays = dict([(p.name, p) for p in particles])
        dummy = get_particle_array_pcisph(name='junk')
        props = list(dummy.properties.keys())
        props += [dict(name=x, stride=y) for x, y in dummy.stride.items()]
        output_props = dummy.output_property_arrays
        constants = [dict(name=x, data=y) for x, y in dummy.constants.items()]
        for fluid in self.fluids:
            pa = particle_arrays[fluid]
            self._ensure_properties(pa, props, clean)
            pa.set_output_arrays(output_props)
            for const in constants:
                pa.add_constant(**const)



from pysph.base.nnps_base import get_number_of_threads, py_flatten, \
        py_unflatten, py_get_valid_cell_index

from pysph.base.nnps_base import NNPSParticleArrayWrapper, CPUDomainManager, \
        DomainManager, Cell, NeighborCache, NNPSBase, NNPS
from pysph.base.linked_list_nnps import LinkedListNNPS
from pysph.base.box_sort_nnps import BoxSortNNPS, DictBoxSortNNPS
from pysph.base.spatial_hash_nnps import SpatialHashNNPS, \
        ExtendedSpatialHashNNPS
from pysph.base.cell_indexing_nnps import CellIndexingNNPS
from pysph.base.z_order_nnps import ZOrderNNPS, ExtendedZOrderNNPS
from pysph.base.stratified_hash_nnps import StratifiedHashNNPS
from pysph.base.stratified_sfc_nnps import StratifiedSFCNNPS
from pysph.base.octree_nnps import OctreeNNPS, CompressedOctreeNNPS



"""Sedov point explosion problem. (7 minutes)

Particles are distributed on concentric circles about the origin with
increasing number of particles with increasing radius. A unit charge
is distributed about the center which gives the initial pressure
disturbance.

"""
# NumPy and standard library imports
import os.path
import numpy

# PySPH base and carray imports
from pysph.base.utils import get_particle_array as gpa
from pysph.solver.application import Application
from pysph.sph.scheme import GasDScheme, SchemeChooser
from pysph.sph.gas_dynamics.psph import PSPHScheme
from pysph.sph.gas_dynamics.tsph import TSPHScheme

# Numerical constants
dim = 2
gamma = 5.0/3.0
gamma1 = gamma - 1.0

# solution parameters
dt = 1e-4
tf = 0.1

# scheme constants
alpha1 = 10.0
alpha2 = 1.0
beta = 2.0
kernel_factor = 1.2


class SedovPointExplosion(Application):
    def create_particles(self):
        fpath = os.path.join(
            os.path.dirname(__file__), 'ndspmhd-sedov-initial-conditions.npz'
        )
        data = numpy.load(fpath)
        x = data['x']
        y = data['y']
        rho = data['rho']
        p = data['p']
        e = data['e'] + 1e-9
        h = data['h']
        m = data['m']

        fluid = gpa(name='fluid', x=x, y=y, rho=rho, p=p, e=e, h=h, m=m)
        self.scheme.setup_properties([fluid])

        # set the initial smoothing length proportional to the particle
        # volume
        fluid.h[:] = kernel_factor * (fluid.m/fluid.rho)**(1./dim)

        print("Sedov's point explosion with %d particles"
              % (fluid.get_number_of_particles()))

        return [fluid,]

    def create_scheme(self):
        mpm = GasDScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            kernel_factor=kernel_factor, alpha1=alpha1, alpha2=alpha2,
            beta=beta, adaptive_h_scheme="mpm",
            update_alpha1=True, update_alpha2=True
        )
        psph = PSPHScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            hfact=kernel_factor
        )

        tsph = TSPHScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            hfact=kernel_factor
        )
        s = SchemeChooser(
            default='mpm', mpm=mpm, psph=psph, tsph=tsph
        )
        return s

    def configure_scheme(self):
        s = self.scheme
        s.configure_solver(
            dt=dt, tf=tf, adaptive_timestep=False, pfreq=25
        )


if __name__ == '__main__':
    app = SedovPointExplosion()
    app.run()



"""3D Dam Break Over a block of text spelling "SPH". (8 hours)

This example demonstrates how to setup a rigid body motion and couple it to the
fluid motion.
"""
from os.path import dirname, join
import numpy as np

from pysph.examples._db_geometry import DamBreak3DGeometry

from pysph.base.kernels import WendlandQuintic
from pysph.base.utils import get_particle_array_rigid_body

from pysph.sph.equation import Group
from pysph.sph.basic_equations import ContinuityEquation, XSPHCorrection
from pysph.sph.wc.basic import TaitEOS, TaitEOSHGCorrection, MomentumEquation

from pysph.solver.application import Application
from pysph.solver.solver import Solver
from pysph.sph.integrator import EPECIntegrator
from pysph.sph.integrator_step import WCSPHStep
from pysph.tools.gmsh import vtk_file_to_points

from pysph.sph.rigid_body import (
    BodyForce, NumberDensity, RigidBodyForceGPUGems,
    RigidBodyMoments, RigidBodyMotion,
    RK2StepRigidBody, PressureRigidBody
)


dim = 3

dt = 1e-5
tf = 2.0

# parameter to chane the resolution
dx = 0.02
nboundary_layers = 3
hdx = 1.2
rho0 = 1000.0


class DamBreak3DSPH(Application):
    def initialize(self):
        self.geom = DamBreak3DGeometry(
            dx=dx, nboundary_layers=nboundary_layers, hdx=hdx, rho0=rho0,
            with_obstacle=False)

    def create_particles(self):
        fluid, boundary = self.geom.create_particles()

        fpath = join(dirname(__file__), 'sph.vtk.gz')
        x, y, z = vtk_file_to_points(fpath, cell_centers=False)
        y -= 0.15
        z += 0.05
        m = np.ones_like(x)*fluid.m[0]
        h = np.ones_like(x)*fluid.h[0]
        rho = np.ones_like(x)*fluid.rho[0]

        obstacle = get_particle_array_rigid_body(
            name='obstacle', x=x, y=y, z=z, m=m, h=h, rho=rho, rho0=rho
        )
        obstacle.total_mass[0] = np.sum(m)
        obstacle.add_property('cs')
        obstacle.add_property('arho')
        obstacle.set_lb_props(list(obstacle.properties.keys()))
        obstacle.set_output_arrays(
            ['x', 'y', 'z', 'u', 'v', 'w', 'fx', 'fy', 'fz',
             'rho', 'm', 'h', 'p', 'tag', 'pid', 'gid']
        )

        boundary.add_property('V')
        boundary.add_property('fx')
        boundary.add_property('fy')
        boundary.add_property('fz')

        return [fluid, boundary, obstacle]

    def create_solver(self):
        kernel = WendlandQuintic(dim=dim)

        integrator = EPECIntegrator(fluid=WCSPHStep(),
                                    obstacle=RK2StepRigidBody(),
                                    boundary=WCSPHStep())
        solver = Solver(kernel=kernel, dim=dim, integrator=integrator,
                        tf=tf, dt=dt, adaptive_timestep=True, n_damp=0)
        return solver

    def create_equations(self):
        co = 10.0 * self.geom.get_max_speed(g=9.81)

        gamma = 7.0
        alpha = 0.5
        beta = 0.0

        equations = [

            Group(equations=[
                    BodyForce(dest='obstacle', sources=None, gz=-9.81),
                    NumberDensity(dest='obstacle', sources=['obstacle']),
                    NumberDensity(dest='boundary', sources=['boundary']),
                    ], ),

            # Equation of state
            Group(equations=[

                    TaitEOS(
                        dest='fluid', sources=None, rho0=rho0, c0=co,
                        gamma=gamma
                    ),
                    TaitEOSHGCorrection(
                        dest='boundary', sources=None, rho0=rho0, c0=co,
                        gamma=gamma
                    ),
                    TaitEOSHGCorrection(
                        dest='obstacle', sources=None, rho0=rho0, c0=co,
                        gamma=gamma
                    ),

                    ], real=False),

            # Continuity, momentum and xsph equations
            Group(equations=[
                    ContinuityEquation(
                        dest='fluid', sources=['fluid', 'boundary', 'obstacle']
                    ),
                    ContinuityEquation(dest='boundary', sources=['fluid']),
                    ContinuityEquation(dest='obstacle', sources=['fluid']),

                    MomentumEquation(dest='fluid',
                                     sources=['fluid', 'boundary'],
                                     alpha=alpha, beta=beta, gz=-9.81, c0=co,
                                     tensile_correction=True),

                    PressureRigidBody(
                        dest='fluid', sources=['obstacle'], rho0=rho0
                    ),

                    XSPHCorrection(dest='fluid', sources=['fluid']),

                    RigidBodyForceGPUGems(
                        dest='obstacle', sources=['boundary'], k=1.0, d=2.0,
                        eta=0.1, kt=0.1
                    ),

                    ]),
            Group(equations=[RigidBodyMoments(dest='obstacle', sources=None)]),
            Group(equations=[RigidBodyMotion(dest='obstacle', sources=None)]),

        ]
        return equations


if __name__ == '__main__':
    app = DamBreak3DSPH()
    app.run()
